from __future__ import absolute_import

import logging
from uuid import uuid4

import six

from omegaml.util import is_dataframe, settings, is_ndarray, extend_instance

logger = logging.getLogger(__name__)


class OmegaModelProxy(object):
    """
    proxy to a remote model in a celery worker

    The proxy provides the same methods as the model but will
    execute the methods using celery tasks and return celery
    AsyncResult objects

    Usage:

        .. code::

            om = Omega()
            # train a model
            # result is AsyncResult, use .get() to return it's result
            result = om.runtime.model('foo').fit('datax', 'datay')
            result.get()

            # predict
            result = om.runtime.model('foo').predict('datax')
            # result is AsyncResult, use .get() to return it's result
            print result.get()
    """
    #     Implementation note:
    #
    #     We decided to implement each method call explicitely in both
    #     this class (mixins) and the celery tasks. While it would be possible to
    #     implement a generic method and task that passes the method and
    #     arguments to be called, maintainability would suffer and the
    #     documentation would become very unspecific. We think it is much
    #     cleaner to have an explicit interface at the chance of missing
    #     features. If need should arise we can still implement a generic
    #     method call.

    def __init__(self, modelname, runtime=None):
        self.modelname = modelname
        self.runtime = runtime
        self.apply_mixins()
        self.apply_require()

    def apply_require(self):
        meta = self.runtime.omega.models.metadata(self.modelname)
        assert meta is not None, "model {self.modelname} does not exist".format(**locals())
        self.runtime.require(meta.attributes.get('require', {}))

    def apply_mixins(self):
        """
        apply mixins in defaults.OMEGA_RUNTIME_MIXINS
        """
        from omegaml import settings
        defaults = settings()
        for mixin in defaults.OMEGA_RUNTIME_MIXINS:
            extend_instance(self, mixin)

    def task(self, name):
        """
        return the task from the runtime with requirements applied
        """
        return self.runtime.task(name)



from multiprocessing import Pool

import math
import os
from itertools import repeat

from omegaml.mongoshim import MongoClient

# single instance multiprocessing pool
pool = None


def dfchunker(df, size=10000):
    """ chunk a dataframe as in iterator """
    return (df.iloc[pos:pos + size].copy() for pos in range(0, len(df), size))


def insert_chunk(job):
    """
    insert one chunk of data

    :param job: the (dataframe, mongo_url, collection_name) tuple. mongo_url
                should include the database name, as the collection is taken
                from the default database of the connection.
    """
    sdf, mongo_url, collection_name = job
    client = MongoClient(mongo_url, authSource='admin')
    db = client.get_database()
    collection = db[collection_name]
    result = collection.insert_many(sdf.to_dict(orient='records'))
    client.close()
    return mongo_url, db.name, collection_name, len(result.inserted_ids)


def fast_insert(df, omstore, name, chunk_size=int(1e4)):
    """
    fast insert of dataframe to mongodb

    Depending on size use single-process or multiprocessing. Typically
    multiprocessing is faster on datasets with > 10'000 data elements
    (rows x columns). Note this may max out your CPU and may use
    processor count * chunksize of additional memory. The chunksize is
    set to 10'000. The processor count is the default used by multiprocessing,
    typically the number of CPUs reported by the operating system.

    :param df: dataframe
    :param omstore: the OmegaStore to use. will be used to get the mongo_url
    :param name: the dataset name in OmegaStore to use. will be used to get the
    collection name from the omstore
    """
    global pool
    if len(df) * len(df.columns) > chunk_size:
        mongo_url = omstore.mongo_url
        collection_name = omstore.collection(name).name
        # we crossed upper limits of single threaded processing, use a Pool
        # use the cached pool
        cores = max(1, math.ceil(os.cpu_count() / 2))
        pool = pool or Pool(processes=cores)
        jobs = zip(dfchunker(df, size=chunk_size),
                   repeat(mongo_url), repeat(collection_name))
        pool.map(insert_chunk, (job for job in jobs))
    else:
        # still within bounds for single threaded inserts
        omstore.collection(name).insert_many(df.to_dict(orient='records'))



from omegaml.backends.basedata import BaseDataBackend

class TFDatasetBackend(BaseDataBackend):
    KIND = 'tf.dataset'

    @classmethod
    def supports(self, obj, name, **kwargs):
        import tensorflow as tf
        return isinstance(obj, )

    def put(self, obj, name, attributes=None, **kwargs):
        pass



"""
omega runtime model tasks 
"""
from __future__ import absolute_import

import os
import sys

import datetime
from celery import shared_task
from celery.signals import worker_process_init

from omegaml.celery_util import OmegamlTask, sanitized

if os.path.basename(sys.argv[0]) == 'celery':
    try:
        # ensure tensorflow is loaded -- this avoids AttributeError for tf.estimator later
        # we do this here instead of in tfestimatormodel to avoid loading tensorflow if it
        # is not needed. loading means this happens at runtime worker startup
        import tensorflow as tf

        tf.version
    except:
        pass


@shared_task(base=OmegamlTask, bind=True)
def omega_predict(self, modelname, Xname, rName=None, pure_python=True, **kwargs):
    result = self.get_delegate(modelname).predict(*self.delegate_args, **self.delegate_kwargs)
    return sanitized(result)


@shared_task(base=OmegamlTask, bind=True)
def omega_predict_proba(self, modelname, Xname, rName=None, pure_python=True,
                        **kwargs):
    result = self.get_delegate(modelname).predict_proba(*self.delegate_args, **self.delegate_kwargs)
    return sanitized(result)


@shared_task(base=OmegamlTask, bind=True)
def omega_fit(self, modelname, Xname, Yname=None, pure_python=True, **kwargs):
    result = self.get_delegate(modelname).fit(*self.delegate_args, **self.delegate_kwargs)
    return sanitized(result)


@shared_task(base=OmegamlTask, bind=True)
def omega_partial_fit(self,
                      modelname, Xname, Yname=None, pure_python=True, **kwargs):
    result = self.get_delegate(modelname).partial_fit(*self.delegate_args, **self.delegate_kwargs)
    return sanitized(result)


@shared_task(base=OmegamlTask, bind=True)
def omega_score(self, modelname, Xname, Yname, rName=True, pure_python=True,
                **kwargs):
    result = self.get_delegate(modelname).score(*self.delegate_args, **self.delegate_kwargs)
    return sanitized(result)


@shared_task(base=OmegamlTask, bind=True)
def omega_fit_transform(self, modelname, Xname, Yname=None, rName=None,
                        pure_python=True, **kwargs):
    result = self.get_delegate(modelname).fit_transform(*self.delegate_args, **self.delegate_kwargs)
    return sanitized(result)


@shared_task(base=OmegamlTask, bind=True)
def omega_transform(self, modelname, Xname, rName=None, **kwargs):
    result = self.get_delegate(modelname).transform(*self.delegate_args, **self.delegate_kwargs)
    return sanitized(result)


@shared_task(base=OmegamlTask, bind=True)
def omega_decision_function(self, modelname, Xname, rName=None, **kwargs):
    result = self.get_delegate(modelname).decision_function(*self.delegate_args, **self.delegate_kwargs)
    return sanitized(result)


@shared_task(base=OmegamlTask, bind=True)
def omega_gridsearch(self, modelname, Xname, Yname, parameters=None, **kwargs):
    result = self.get_delegate(modelname).gridsearch(*self.delegate_args, **self.delegate_kwargs)
    return sanitized(result)


@shared_task(base=OmegamlTask, bind=True)
def omega_settings(self, *args, **kwargs):
    if os.environ.get('OMEGA_DEBUG'):
        defaults = self.om.defaults
        return {k: getattr(defaults, k, '')
                for k in dir(defaults) if k and k.isupper()}
    return {'error': 'settings dump is disabled'}


@shared_task(base=OmegamlTask, bind=True)
def omega_ping(task, *args, **kwargs):
    import socket
    hostname = task.request.hostname or socket.gethostname()
    # resolve standard kwargs
    om = task.om
    args = task.delegate_args
    kwargs = task.delegate_kwargs
    kwargs.pop('pure_python', None)
    # return ping
    return {
        'message': 'ping return message',
        'time': datetime.datetime.now().isoformat(),
        'args': args,
        'kwargs': kwargs,
        'worker': hostname,
    }


@worker_process_init.connect
def fix_multiprocessing(**kwargs):
    # allow celery to start sub processes
    # this is required for sklearn joblib unpickle support
    # issue see https://github.com/celery/billiard/issues/168
    # fix source https://github.com/celery/celery/issues/1709
    from multiprocessing import current_process
    try:
        current_process()._config
    except AttributeError:
        current_process()._config = {'semprefix': '/mp'}



# this will be executed on jupyter notebook / ipython startup
import sys
import os
base_dir = os.environ.get('OMEGA_ROOT')
if not base_dir:
    base_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..')
sys.path.insert(0, base_dir)
import omegaml as om
print("omegaml initializing from {}".format(__file__))
print(om.defaults.OMEGA_CONFIG_FILE)



# http://ipython.readthedocs.io/en/5.x/config/intro.html#profiles
# http://ipython.readthedocs.io/en/5.x/config/intro.html#example-config-file
# note in the link above it says "you can also keep a profile in the
# current working directory". so put this file in the cwd and it will actually
# execute on kernel starts

c = get_config()

c.InteractiveShellApp.exec_files = [
    'ipystart.py'
]
print("omegaml: ipython initialized from {}".format(__file__))



from sklearn.pipeline import Pipeline


class OnlinePipeline(Pipeline):
    """
    A Pipeline that supports online training through partial_fit

    Consider:
        https://github.com/scikit-learn/scikit-learn/issues/3299
        http://scikit-learn.org/stable/modules/scaling_strategies.html#incremental-learning
    """

    def __init__(self, steps, safe=False):
        super(OnlinePipeline, self).__init__(steps)
        self.safe = safe

    def partial_fit(self, X, y=None):
        """
        Apply partial_fit to all steps.

        Note this will fail if a step's estimator does not
        support partial_fit. You can avoid failing by making
        the OnlinePipeline(safe=True), at the risk of introducing
        unwanted semantic differences

        :param X: the X
        :param y: the y
        :return: self
        """
        for i, step in enumerate(self.steps):
            name, est = step
            if self.safe and not hasattr(est, 'partial_fit'):
                # we allow to be safe on all but the last step
                # i.e. only transformers are ok to not have a partial_fit
                if i < len(self.steps) - 1:
                    continue
            est.partial_fit(X, y)
            if i < len(self.steps) - 1:
                # pass on results to next step
                X = est.transform(X)
        return self



from flask import Flask
from flask_restplus import Api
from werkzeug.utils import redirect

app = Flask(__name__)
api = Api(app)


@app.route('/docs')
def docs():
    return redirect("https://omegaml.github.io/omegaml/", code=302)



import glob
import os
import tempfile
from shutil import rmtree
from zipfile import ZipFile, ZIP_DEFLATED

import numpy as np
from mongoengine import GridFSProxy

from omegaml.backends import BaseModelBackend


class TensorflowSavedModelPredictor(object):
    """
    A predictor model from a TF SavedModel
    """

    def __init__(self, model_dir):
        from tensorflow.contrib import predictor
        self.predict_fn = predictor.from_saved_model(model_dir)
        self.input_names = list(self.predict_fn.feed_tensors.keys())
        self.output_names = list(self.predict_fn.fetch_tensors.keys())

    def _convert_to_model_input(self, X):
        # coerce input into expected feature mapping
        if len(self.input_names) > 1:
            raise ValueError('multiple inputs not supported')
        model_input = {
            self.input_names[0]: X
        }
        return model_input

    def _convert_to_model_output(self, yhat):
        # coerce output into dict or array-like response
        if len(self.output_names) == 1:
            yhat = yhat[self.output_names[0]]
        return yhat

    def predict(self, X):
        yhat = self.predict_fn(self._convert_to_model_input(X))
        return self._convert_to_model_output(yhat)


class TensorflowSavedModelBackend(BaseModelBackend):
    KIND = 'tf.savedmodel'
    _model_ext = 'tfsm'

    @classmethod
    def supports(self, obj, name, **kwargs):
        import tensorflow as tf
        return isinstance(obj, tf.estimator.Estimator)

    def _package_savedmodel(self, export_base_dir, filename):
        fname = os.path.basename(filename)
        zipfname = os.path.join(self.model_store.tmppath, fname)
        # check if we have an intermediate directory (timestamp)
        # as in export_base_dir/<timestamp>, if so, use this as the base directory
        # see https://www.tensorflow.org/guide/saved_model#perform_the_export
        # we need this check because not all SavedModel exports create a timestamp
        # directory. e.g. keras.save_keras_model() does not, while Estimator.export_saved_model does
        files = glob.glob(os.path.join(export_base_dir, '*'))
        if len(files) == 1:
            export_base_dir = files[0]
        with ZipFile(zipfname, 'w', compression=ZIP_DEFLATED) as zipf:
            for part in glob.glob(os.path.join(export_base_dir, '**'), recursive=True):
                zipf.write(part, os.path.relpath(part, export_base_dir))
        return zipfname

    def _extract_savedmodel(self, packagefname):
        lpath = tempfile.mkdtemp()
        fname = os.path.basename(packagefname)
        mklfname = os.path.join(lpath, fname)
        with ZipFile(packagefname) as zipf:
            zipf.extractall(lpath)
        model = TensorflowSavedModelPredictor(lpath)
        rmtree(lpath)
        return model

    def _make_savedmodel(self, obj, serving_input_receiver_fn=None, strip_default_attrs=None):
        # adapted from https://www.tensorflow.org/guide/saved_model#perform_the_export
        export_dir_base = tempfile.mkdtemp()
        obj.export_savedmodel(export_dir_base,
                              serving_input_receiver_fn=serving_input_receiver_fn,
                              strip_default_attrs=strip_default_attrs)
        return export_dir_base

    def put_model(self, obj, name, attributes=None, serving_input_fn=None,
                  strip_default_attrs=None, **kwargs):
        export_dir_base = self._make_savedmodel(obj, serving_input_receiver_fn=serving_input_fn,
                                                strip_default_attrs=strip_default_attrs)
        zipfname = self._package_savedmodel(export_dir_base, name)
        with open(zipfname, 'rb') as fzip:
            fileid = self.model_store.fs.put(
                fzip, filename=self.model_store._get_obj_store_key(name, self._model_ext))
            gridfile = GridFSProxy(grid_id=fileid,
                                   db_alias='omega',
                                   collection_name=self.model_store.bucket)
        rmtree(export_dir_base)
        os.remove(zipfname)
        return self.model_store._make_metadata(
            name=name,
            prefix=self.model_store.prefix,
            bucket=self.model_store.bucket,
            kind=self.KIND,
            attributes=attributes,
            gridfile=gridfile).save()

    def get_model(self, name, version=-1):
        filename = self.model_store._get_obj_store_key(name, self._model_ext)
        packagefname = os.path.join(self.model_store.tmppath, name)
        outf = self.model_store.fs.get_version(filename, version=version)
        with open(packagefname, 'wb') as zipf:
            zipf.write(outf.read())
        model = self._extract_savedmodel(packagefname)
        os.remove(packagefname)
        return model

    def predict(
            self, modelname, Xname, rName=None, pure_python=True, **kwargs):
        """
        Predict from a SavedModel

        Args:
            modelname:
            Xname:
            rName:
            pure_python:
            kwargs:

        Returns:

        """
        import pandas as pd
        model = self.get_model(modelname)
        X = self.data_store.get(Xname)
        result = model.predict(X)
        if pure_python:
            result = result.tolist()
        if rName:
            result = self.data_store.put(result, rName)
        return result

    def fit(self, modelname, Xname, Yname=None, pure_python=True, tpu_specs=None, **kwargs):
        raise ValueError('cannot fit a saved model')


class ServingInput(object):
    # FIXME this is not working yet
    def __init__(self, model=None, features=None, like=None, shape=None, dtype=None,
                 batchsize=1, from_keras=False):
        """
        Helper to create serving_input_fn

        Uses tf.build_raw_serving_input_receiver_fn to build a ServingInputReceiver
        from the given inputs

        Usage:
            # use existing ndarray e.g. training or test data to specify a single input feature
            ServingInput(features=['x'], like=ndarray)

            # specify the dtype and shape explicitely
            ServingInput(features=['x'], shape=(1, 28, 28))

            # use multiple features
            ServingInput(features={'f1': tf.Feature(...))

            # for tf.keras models turned estimator, specify from_keras
            # to ensure the input features are renamed correctly.
            ServingInput(features=['x'], like=ndarray, from_keras=True)

        Args:
            model:
            features:
            like:
            shape:
            dtype:
            batchsize:
            from_keras:
        """
        self.model = model
        self.features = features or ['X']
        self.like = like
        self.shape = shape
        self.dtype = dtype
        self.batchsize = batchsize
        self.from_keras = from_keras

    def build(self):
        if isinstance(self.features, dict):
            input_fn = self.from_features()
        elif isinstance(self.like, np.ndarray):
            shape = tuple((self.batchsize, *self.like.shape[1:])) # assume (rows, *cols)
            input_fn = self.from_ndarray(shape, self.like.dtype)
        elif isinstance(self.shape, (list,tuple,np.ndarray)):
            input_fn = self.from_ndarray(self.shape, self.dtype)
        return input_fn

    def __call__(self):
        input_fn = self.build()
        return input_fn()

    def from_features(self):
        import tensorflow as tf
        input_fn = tf.estimator.export.build_raw_serving_input_receiver_fn(
            self.features,
            default_batch_size=self.batchsize
        )
        return input_fn

    def from_ndarray(self, shape, dtype):
        import tensorflow as tf
        if self.from_keras:
            input_layer_name = '{}_input'.format(self.features[0])
        else:
            input_layer_name = self.features[0]
        features = {
           input_layer_name: tf.placeholder(dtype=dtype, shape=shape, )
        }
        input_fn = tf.estimator.export.build_raw_serving_input_receiver_fn(
            features,
            default_batch_size=None
        )
        return input_fn

    def from_dataframe(self, columns, input_layer_name='X',
                       batch_size=1, dtype=np.float32):
        def serving_input_fn():
            import tensorflow as tf
            ndim = len(columns)
            X_name = '{}_input'.format(input_layer_name)
            placeholder = tf.placeholder(dtype=np.float32,
                                         shape=(batch_size, ndim),
                                         name=X_name)
            receiver_tensors = {X_name: placeholder}
            features = {X_name: placeholder}
            return tf.estimator.export.ServingInputReceiver(features, receiver_tensors)








from __future__ import absolute_import

from omegaml.util import PickableCollection
from pymongo.collection import Collection

from omegaml.store import qops
from omegaml.store.query import Filter


class FilteredCollection(Collection):

    """
    A permanently filtered collection

    Supports all methods as a Collection does, however any filter or query
    argument is permanently set at instantiation

        fcoll = FilteredCollection(collection, query={ expression })

    Any subsequent operation will automatically apply the query expression. 

    Note that v.v. a Collection and all methods that accept a filter as their first  
    argument have a changed signature - the filter argument is optional
    with all FilteredCollection methods, as the filter is set at instantiation.

        Example:

            # in pymongo

            filter = {expression}
            coll.find_one_and_replace(filter, replace)

            # FilteredCollection

            coll = FilteredCollection(query={expression})
            coll.find_one_and_replace(replace, filter=None)

    This is so that calls to a FilteredCollection feel more natural, as opposed
    to specifying an empty filter argument on every call. Still, an additional
    filter can be specified on every method that accepts the filter= optional
    argument:  

            # temporarily add another filter

            coll.find_one_and_replace(replace, filter={expression})

    Here expression will only apply to this particular method call. The
    global filter set by query= is unchanged.

    If no expression is given, the empty expression {} is assumed. To change
    the expression for the set fcoll.query = { expression }  
    """

    def __init__(self, collection, query=None, projection=None, **kwargs):
        if isinstance(collection, (Collection, PickableCollection)):
            database = collection.database
            name = collection.name
        else:
            raise ValueError('collection should be a pymongo.Collection')
        query = query or {}
        super(FilteredCollection, self).__init__(
            database, name, create=False, **kwargs)
        self.query = Filter(self, **query).query
        self.projection = projection
    def replace_one(self, replacement, filter=None, upsert=False,
                    bypass_document_validation=False):
        query = dict(self.query)
        query.update(filter or {})
        kwargs = dict(query, replacement, upsert=upsert,
                      bypass_document_validation=bypass_document_validation)
        return super(FilteredCollection, self).replace_one(**kwargs)
    def update_one(self, update, filter=None, upsert=False,
                   bypass_document_validation=False):
        query = dict(self.query)
        query.update(filter or {})
        kwargs = dict(query, update, upsert=upsert,
                      bypass_document_validation=bypass_document_validation)
        return super(FilteredCollection, self).update_one(**kwargs)
    def update_many(self, update, filter=None, upsert=False,
                    bypass_document_validation=False):
        query = dict(self.query)
        query.update(filter or {})
        kwargs = dict(query, update, upsert=upsert,
                      bypass_document_validation=bypass_document_validation)
        return super(FilteredCollection, self).update_many(**kwargs)
    def delete_one(self, filter=None):
        query = dict(self.query)
        query.update(filter or {})
        return super(FilteredCollection, self).delete_one(query)
    def delete_many(self, filter=None):
        query = dict(self.query)
        query.update(filter or {})
        return super(FilteredCollection, self).delete_many(query)
    def aggregate(self, pipeline, filter=None, **kwargs):
        query = dict(self.query)
        query.update(filter or {})
        pipeline.insert(0, qops.MATCH(query))
        return super(FilteredCollection, self).aggregate(pipeline, **kwargs)
    def find(self, filter=None, **kwargs):
        query = dict(self.query)
        query.update(filter or {})
        return super(FilteredCollection, self).find(filter=query, **kwargs)
    def find_one(self, filter=None, *args, **kwargs):
        query = dict(self.query)
        query.update(filter or {})
        return super(FilteredCollection, self).find_one(query, *args, **kwargs)
    def find_one_and_delete(self, filter=None, **kwargs):
        query = dict(self.query)
        query.update(filter or {})
        return super(FilteredCollection, self).find_one_and_delete(query,
                                                                   **kwargs)
    def find_one_and_replace(self, replacement, filter=None, **kwargs):
        query = dict(self.query)
        query.update(filter or {})
        return super(FilteredCollection, self).find_one_and_replace(query,
                                                                    replacement,
                                                                    **kwargs)
    def find_one_and_update(self, update, filter=None, **kwargs):
        query = dict(self.query)
        query.update(filter or {})
        return super(FilteredCollection, self).find_one_and_update(query,
                                                                   update,
                                                                   **kwargs)
    def count(self, filter=None, **kwargs):
        query = dict(self.query)
        query.update(filter or {})
        return super(FilteredCollection, self).count(filter=query, **kwargs)
    def distinct(self, key, filter=None, **kwargs):
        query = dict(self.query)
        query.update(filter or {})
        return super(FilteredCollection, self).distinct(key, filter=query,
                                                        **kwargs)
    def group(self, key, initial, reduce, condition=None, **kwargs):
        condition = dict(self.query)
        condition.update(condition or {})
        return super(FilteredCollection, self).group(key, condition, initial,
                                                     reduce,
                                                     **kwargs)
    def map_reduce(self, m, r, out, full_response=False, query=None, **kwargs):
        _query = dict(self.query)
        _query.update(query or {})
        return super(FilteredCollection, self).map_reduce(m, r, out,
                                                          full_response=False,
                                                          query=_query, **kwargs)
    def inline_map_reduce(self, m, r, full_response=False,
                          query=None, **kwargs):
        _query = dict(self.query)
        _query.update(query or {})
        return super(FilteredCollection, self).inline_map_reduce(m, r,
                                                                 full_response=False, query=_query, **kwargs)
    def insert(self, *args, **kwargs):
        raise NotImplementedError(
            "deprecated in Collection and not implemented in FilteredCollection")
    def update(self, *args, **kwargs):
        raise NotImplementedError(
            "deprecated in Collection and not implemented in FilteredCollection")
    def remove(self, *args, **kwargs):
        raise NotImplementedError(
            "deprecated in Collection and not implemented in FilteredCollection")
    def find_and_modify(self, *args, **kwargs):
        raise NotImplementedError(
            "deprecated in Collection and not implemented in FilteredCollection")
    def ensure_index(self, *args, **kwargs):
        raise NotImplementedError(
            "deprecated in Collection and not implemented in FilteredCollection")
    def save(self, *args, **kwargs):
        raise NotImplementedError(
            "deprecated in Collection and not implemented in FilteredCollection")



from unittest import TestCase

import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression

from omegaml import Omega
from omegaml.client.auth import OmegaRestApiAuth
from omegaml.restapi.app import app
from omegaml.restapi.tests.test_api import OmegaRestApiTests
from omegaml.restapi.tests.util import RequestsLikeTestClient


class OmegaRestApiTestsWithBuckets(OmegaRestApiTests):
    def setUp(self):
        self.client = RequestsLikeTestClient(app)
        self.om = Omega()['mybucket']
        self.auth = OmegaRestApiAuth('user', 'pass')
        self.clean()
        self.clean('mybucket')

    @property
    def _headers(self):
        return dict(bucket='mybucket')

    def test_default_bucket_fails(self):
        # put a model in the default bucket
        om = Omega()
        X = np.arange(10).reshape(-1, 1)
        y = X * 2
        # train model locally
        clf = LinearRegression()
        clf.fit(X, y)
        result = clf.predict(X)
        # store model in om
        om.models.put(clf, 'regression')
        resp = self.client.put('/api/v1/model/regression/predict', json={
            'columns': ['v'],
            'data': [dict(v=[5])],
        }, auth=self.auth, headers=self._headers)
        # we expect an error because the model does not exist in the default bucket
        self.assertEqual(resp.status_code, 500)
        # see if we can get it to predict with the correct bucket (all other tests do this)
        # -- note we simply remove the the 'bucket' header which reverts to the default
        resp = self.client.put('/api/v1/model/regression/predict', json={
            'columns': ['v'],
            'data': dict(v=[5])
        }, auth=self.auth)
        self.assertEqual(resp.status_code, 200)



from __future__ import absolute_import

from celery import Celery

from omegaml.runtimes.jobproxy import OmegaJobProxy
from omegaml.runtimes.scriptproxy import OmegaScriptProxy
from omegaml.util import settings
import logging

logger = logging.getLogger(__name__)


class CeleryTask(object):
    """
    A thin wrapper for a Celery.Task object

    This is so that we can collect common delay arguments on the
    .task() call
    """

    def __init__(self, task, kwargs):
        """

        Args:
            task (Celery.Task): the celery task object
            kwargs (dict): optional, the kwargs to pass to apply_async
        """
        self.task = task
        self.kwargs = kwargs

    def _apply_kwargs(self, task_kwargs, celery_kwargs):
        # update task_kwargs from runtime's passed on kwargs
        # update celery_kwargs to match celery routing semantics
        task_kwargs.update(self.kwargs.get('task', {}))
        celery_kwargs.update(self.kwargs.get('routing', {}))
        if 'label' in celery_kwargs:
            celery_kwargs['queue'] = celery_kwargs['label']
            del celery_kwargs['label']

    def apply_async(self, args=None, kwargs=None, **celery_kwargs):
        """

        Args:
            args (tuple): the task args
            kwargs (dict): the task kwargs
            celery_kwargs (dict): apply_async kwargs, e.g. routing

        Returns:
            AsyncResult
        """
        self._apply_kwargs(kwargs, celery_kwargs)
        return self.task.apply_async(args=args, kwargs=kwargs, **celery_kwargs)

    def delay(self, *args, **kwargs):
        """
        submit the task with args and kwargs to pass on

        This calls task.apply_async and passes on the self.kwargs.
        """
        return self.apply_async(args=args, kwargs=kwargs)

    def run(self, *args, **kwargs):
        return self.delay(*args, **kwargs)


class OmegaRuntime(object):
    """
    omegaml compute cluster gateway
    """

    def __init__(self, omega, bucket=None, defaults=None, celeryconf=None):
        self.omega = omega
        defaults = defaults or settings()
        self.bucket = bucket
        self.pure_python = getattr(defaults, 'OMEGA_FORCE_PYTHON_CLIENT', False)
        self.pure_python = self.pure_python or self._client_is_pure_python()
        # initialize celery as a runtimes
        taskpkgs = defaults.OMEGA_CELERY_IMPORTS
        celeryconf = celeryconf or defaults.OMEGA_CELERY_CONFIG
        # ensure we use current value
        celeryconf['CELERY_ALWAYS_EAGER'] = bool(defaults.OMEGA_LOCAL_RUNTIME)
        self.celeryapp = Celery('omegaml')
        self.celeryapp.config_from_object(celeryconf)
        # needed to get it to actually load the tasks
        # https://stackoverflow.com/a/35735471
        self.celeryapp.autodiscover_tasks(taskpkgs, force=True)
        self.celeryapp.finalize()
        # temporary requirements, use .require() to set
        self._require_kwargs = dict(task={}, routing={})
        # fixed default arguments, use .require(always=True) to set
        self._task_default_kwargs = dict(task={}, routing={})
        # default routing label
        self._default_label = self.celeryapp.conf.get('CELERY_DEFAULT_QUEUE')

    def __repr__(self):
        return 'OmegaRuntime({})'.format(self.omega.__repr__())

    @property
    def _common_kwargs(self):
        common = dict(self._task_default_kwargs)
        common['task'].update(pure_python=self.pure_python, __bucket=self.bucket)
        common['routing'].update(self._require_kwargs['routing'])
        return common

    def _client_is_pure_python(self):
        try:
            import pandas as pd
            import numpy as np
            import sklearn
        except Exception as e:
            logging.getLogger().info(e)
            return True
        else:
            return False

    def require(self, label=None, always=False, **kwargs):
        """
        specify requirements for the task execution

        Use this to specify resource or routing requirements on the next task
        call sent to the runtime. Any requirements will be reset after the
        call has been submitted.

        Args:
            always (bool): if True requirements will persist across task calls. defaults to False
            label (str): the label required by the worker to have a runtime task dispatched to it
            kwargs: requirements specification that the runtime understands

        Usage:
            om.runtime.require(label='gpu').model('foo').fit(...)

        Returns:
            self
        """
        kwargs.update({'label': label or self._default_label})
        if always:
            self._task_default_kwargs['routing'].update(kwargs)
        else:
            self._require_kwargs['routing'].update(kwargs)
        return self

    def model(self, modelname, require=None):
        """
        return a model for remote execution

        Args:
            require (dict): routing requirements for this job
        """
        from omegaml.runtimes.modelproxy import OmegaModelProxy
        self.require(**require) if require else None
        return OmegaModelProxy(modelname, runtime=self)

    def job(self, jobname, require=None):
        """
        return a job for remote exeuction

        Args:
            require (dict): routing requirements for this job
        """
        self.require(**require) if require else None
        return OmegaJobProxy(jobname, runtime=self)

    def script(self, scriptname, require=None):
        """
        return a script for remote execution

        Args:
            require (dict): routing requirements for this job
        """
        self.require(**require) if require else None
        return OmegaScriptProxy(scriptname, runtime=self)

    def task(self, name):
        """
        retrieve the task function from the celery instance

        Args:
            kwargs (dict): routing keywords to CeleryTask.apply_async
        """
        taskfn = self.celeryapp.tasks.get(name)
        assert taskfn is not None, "cannot find task {name} in Celery runtime".format(**locals())
        task = CeleryTask(taskfn, self._common_kwargs)
        self._require_kwargs = dict(routing={}, task={})
        return task

    def settings(self, require=None):
        """
        return the runtimes's cluster settings
        """
        self.require(**require) if require else None
        return self.task('omegaml.tasks.omega_settings').delay().get()

    def ping(self, require=None, *args, **kwargs):
        """
        ping the runtimes

        Args:
            require (dict): routing requirements for this job
            args (tuple): task args
            kwargs (dict): task kwargs
        """
        self.require(**require) if require else None
        return self.task('omegaml.tasks.omega_ping').delay(*args, **kwargs).get()



from collections import defaultdict
from unittest import TestCase

from sklearn.linear_model import LinearRegression

from omegaml import Omega
from omegaml.client.cli import main
from omegaml.tests.util import OmegaTestMixin


class ClientCliTest(OmegaTestMixin, TestCase):
    def setUp(self):
        self.om = Omega()
        self.clean()

    def test_cli_models_list(self):
        # start with an empty list
        argv = 'models list'.split(' ')
        parser = main(argv=argv, logger=get_test_logger())
        expected = ([],)
        data = parser.logger.data['info']
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0], expected)
        # add a model, see that we get it back
        reg = LinearRegression()
        self.om.models.put(reg, 'reg')
        parser = main(argv=argv, logger=get_test_logger())
        expected = (['reg'],)
        data = parser.logger.data['info']
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0], expected)
        # check we can get back a Metadata object
        argv = 'models list --raw'.split(' ')
        parser = main(argv=argv, logger=get_test_logger())
        data = parser.logger.data['info']
        self.assertTrue(str(data[0][0][0]).startswith('Metadata('))


def get_test_logger():
    class TestLogger:
        data = defaultdict(list)

        def info(self, *args):
            self.data['info'].append(args)

        def error(self, *args):
            self.data['error'].append(args)

        def debug(self, *args):
            self.data['debug'].append(args)

    return TestLogger()



import os

import six
from mongoengine import GridFSProxy

from omegaml.backends.basedata import BaseDataBackend
from omegaml.backends.package.packager import build_sdist, install_and_import, load_from_path


class PythonPackageData(BaseDataBackend):
    """
    Backend to support custom scripts deployment to runtimes cluster
    """
    KIND = 'python.package'

    @classmethod
    def supports(self, obj, name, **kwargs):
        return isinstance(obj, six.string_types) and obj.startswith('pkg://')

    def put(self, obj, name, attributes=None, **kwargs):
        """
        save a python package

        :param obj: full path to package file or directory, sytanx as
                    pkg://path/to/dist.tar.gz or pkg://path/to/setup.py
        :param name: name of package. must be the actual name of the package
                     as specified in setup.py
        :return: the Metadata object
        """
        pkgsrc = pkgdist = obj.split('//')[1]
        if not 'tar.gz' in os.path.basename(pkgdist):
            distdir = os.path.join(pkgsrc, 'dist')
            sdist = build_sdist(pkgsrc, distdir)
            version = sdist.metadata.version
            pkgname = sdist.metadata.name
            pkgdist = os.path.join(distdir, '{pkgname}-{version}.tar.gz'.format(**locals()))
        with open(pkgdist, 'rb') as fzip:
            fileid = self.data_store.fs.put(
                fzip, filename=self.data_store._get_obj_store_key(name, 'pkg'))
            gridfile = GridFSProxy(grid_id=fileid,
                                   db_alias='omega',
                                   collection_name=self.data_store.bucket)
        return self.data_store._make_metadata(
            name=name,
            prefix=self.data_store.prefix,
            bucket=self.data_store.bucket,
            kind=PythonPackageData.KIND,
            attributes=attributes,
            gridfile=gridfile).save()

    def get(self, name, **kwargs):
        """
        Load package from store, install it locally and load.

        :param name: the name of the package
        :param kwargs:
        :return: the loaded module
        """
        packagefname = '{}.tar.gz'.format(os.path.join(self.data_store.tmppath, name))
        dstdir = self.packages_path
        if not os.path.exists(os.path.join(dstdir, name)):
            filename = self.data_store._get_obj_store_key(name, '.pkg')
            outf = self.data_store.fs.get_version(filename)
            with open(packagefname, 'wb') as pkgf:
                pkgf.write(outf.read())
            mod = install_and_import(packagefname, name, dstdir)
        else:
            mod = load_from_path(name, dstdir)
        return mod

    @property
    def packages_path(self):
        return os.path.join(self.data_store.tmppath, 'packages')


from __future__ import absolute_import

import logging

from omegaml.util import load_class, settings
import omegaml.defaults as _base_config

logger = logging.getLogger(__file__)

try:
    from omegaee import omega as _omega
except Exception as e:
    from omegaml import omega as _omega
except:
    pass

# link implementation
setup = _omega.setup
version = _omega.version
get_omega_for_task = _omega.get_omega_for_task
Omega = _omega.Omega
# setup API
#: the OmegaStore for datasets
datasets = _omega.OmegaDeferredInstance(_omega._om, 'datasets')
#: the OmegaStore for models
models = _omega.OmegaDeferredInstance(_omega._om, 'models')
#: the jobs API
jobs = _omega.OmegaDeferredInstance(_omega._om, 'jobs')
#: the OmegaStore for lambda scripts
scripts = _omega.OmegaDeferredInstance(_omega._om, 'scripts')
#: the OmegaRuntime for cluster execution
runtime = _omega.OmegaDeferredInstance(_omega._om, 'runtime')
#: the OmegaSimpleLogger for easy log access
logger = _omega.OmegaDeferredInstance(_omega._om, 'logger')
#: the settings object
defaults = _omega.OmegaDeferredInstance(_omega._om, 'defaults')




from unittest import TestCase

from omegaml import Omega
from omegaml.backends.npndarray import NumpyNDArrayBackend

import numpy as np


class NumpyNDArrayBackendTests(TestCase):
    def setUp(self):
        om = self.om = Omega()
        om.datasets.register_backend(NumpyNDArrayBackend.KIND, NumpyNDArrayBackend)

    def test_store_ndarray(self):
        om = self.om
        arr = np.random.randint(0, 255, (768, 1204), dtype=np.uint8)
        # store
        meta = om.datasets.put(arr, 'ndarray-test')
        self.assertIsNotNone(meta)
        self.assertEqual(meta.kind, 'ndarray.bin')
        self.assertEqual(meta.name, 'ndarray-test')
        self.assertIn('dtype', meta.kind_meta)
        self.assertIn('shape', meta.kind_meta)
        # get back
        data = om.datasets.get('ndarray-test')
        self.assertIsNotNone(data)
        self.assertEqual(data.shape, arr.shape)
        self.assertEqual(data.dtype, arr.dtype)



"""
omega runtime script tasks
"""
from __future__ import absolute_import

import json

import datetime
from celery import shared_task

from omegaml.backends.package import PythonPackageData
from omegaml.celery_util import OmegamlTask


class NotebookTask(OmegamlTask):
    abstract = True

    def on_success(self, retval, task_id, *args, **kwargs):
        om = self.om
        args, kwargs = args[0:2]
        scriptname = args[0]
        meta = om.scripts.metadata(scriptname)
        attrs = meta.attributes
        attrs['state'] = 'SUCCESS'
        attrs['task_id'] = task_id
        meta.kind = PythonPackageData.KIND

        if not kwargs:
            pass
        else:
            attrs['last_run_time'] = kwargs.get('run_at')
            attrs['next_run_time'] = kwargs.get('next_run_time')

        meta.attributes = attrs
        meta.save()

    def on_failure(self, retval, task_id, *args, **kwargs):
        print("*** NotebookTask.on_failure (retval, task_id, args, kwargs)",
              retval, task_id, args, kwargs)
        om = self.om
        args, kwargs = args[0:2]
        scriptname = args[0]
        meta = om.scripts.metadata(scriptname)
        attrs = meta.attributes
        attrs['state'] = 'FAILURE'
        attrs['task_id'] = task_id
        meta.kind = PythonPackageData.KIND

        if not kwargs:
            pass
        else:
            attrs['last_run_time'] = kwargs.get('run_at')
            attrs['next_run_time'] = kwargs.get('next_run_time')

        meta.attributes = attrs
        meta.save()


@shared_task(bind=True, base=NotebookTask)
def run_omega_script(self, scriptname, **kwargs):
    """
    runs omegaml job
    """
    SERIALIZER = {
        'json': json.dumps,
        'python': lambda v: v,
    }
    format = kwargs.get('__format') or 'json'
    mod = self.om.scripts.get(scriptname)
    dtstart = datetime.datetime.now()
    try:
        result = mod.run(self.om, **self.delegate_kwargs)
    except Exception as e:
        result = str(e)
    dtend = datetime.datetime.now()
    duration = dtend - dtstart

    data = {
        'script': scriptname,
        'kwargs': self.delegate_kwargs,
        'result': result,
        'runtimes': float(duration.seconds) + duration.microseconds / float(1e6),
        'started': dtstart.isoformat(),
    }
    return SERIALIZER[format](data)



from behave import when, then
from time import sleep

from omegaml.tests.features.util import Notebook


@when(u'we open jupyter')
def open_jupyter(ctx):
    br = ctx.browser
    br.visit(ctx.feature.jynb_url)
    nb = Notebook(br)
    if br.is_text_present('JupyterHub', wait_time=10):
        login_required = br.is_element_present_by_id('username_input', wait_time=30)
        if login_required:
            nb.login_hub()
    else:
        # fallback to juypter notebook
        login_required = br.is_text_present('Password', wait_time=2)
        login_required |= br.is_text_present('token', wait_time=2)
        if login_required:
            nb.login_nb()
    nb.jupyter_home


@when(u'we create a notebook')
def create_notebook(ctx):
    br = ctx.browser
    nb = Notebook(br)
    nb.create_notebook()
    nb.save_notebook()
    assert not br.is_text_present('error while saving')
    # test code execution
    code = """
    print('hello')
    """.strip()
    nb.current_cell_exec(code)
    sleep(1)
    assert nb.current_cell_output() == 'hello'
    nb.save_notebook()
    assert not br.is_text_present('error while saving')


@when(u'we restart the notebook')
def restart_kernel(ctx):
    br = ctx.browser
    nb = Notebook(br)
    nb.restart(wait=True)


@when(u'we create a folder')
def create_folder(ctx):
    br = ctx.browser
    nb = Notebook(br)
    nb.create_folder()
    nb.open_folder('Untitled Folder')


@then(u'we can list datasets in omegaml')
def list_datasets(ctx):
    # test omegaml functionality
    br = ctx.browser
    nb = Notebook(br)
    code = """
    import omegaml as om
    om.datasets.put(['sample'], 'sample', append=False)
    om.datasets.list('sample')
    """.strip()
    nb.new_cell_exec(code)
    sleep(10)
    current = nb.current_cell_output()
    expected = "['sample']"
    assert current == expected, "Expected {expected}, got {current}".format(**locals())


@then(u'we can add a notebook in the folder')
def add_notebook_in_folder(ctx):
    br = ctx.browser
    br.visit(ctx.feature.jynb_url)
    nb = Notebook(br)
    nb.jupyter_home
    nb.open_folder('Untitled Folder')
    nb.create_notebook()
    nb.last_notebook
    assert not br.is_text_present('No such directory')



import datetime
import nbformat
from cron_descriptor import get_description
from omegaml.client.docoptparser import CommandBase

from omegaml.client.util import get_omega


class JobsCommandBase(CommandBase):
    """
    Usage:
      om jobs list [<pattern>] [--raw] [options]
      om jobs put <path> <name> [options]
      om jobs get <name> <path> [options]
      om jobs metadata <name> [options]
      om jobs schedule <name> [show|delete|<interval>] [options]
      om jobs status <name>

    Options:
      --cron <spec>       the cron spec, use https://crontab.guru/
      --weekday <wday>    a day number 0-6 (0=Sunday)
      --monthday <mday>   a day of the month 1-31
      --month <month>     a month number 1-12
      --at <hh:mm>        the time (same as --hour hh --minute mm)
      --hour <hour>       the hour 0-23
      --minute <minute>   the minute 0-59
      --next <n>          show next n triggers according to interval

    Description:
        Specify the schedule either as

        * a natural language-like text, with any time components separated
          by comma

          om jobs schedule myjob "every 5 minutes, on fridays, in april"
          om jobs schedule myjob "at 6:00, on fridays"
          om jobs schedule myjob "at 6:00/10:00, on fridays"
          om jobs schedule myjob "every 2nd hour, every 15 minutes, weekdays"


    """
    command = 'jobs'

    def list(self):
        om = get_omega(self.args)
        raw = self.args.get('--raw', False)
        pattern = self.args.get('<pattern>')
        entries = om.jobs.list(jobfilter=pattern, raw=raw)
        self.logger.info(entries)

    def put(self):
        from nbformat import read as nbread

        om = get_omega(self.args)
        local = self.args['<path>']
        name = self.args['<name>']
        with open(local, 'rb') as fin:
            nb = nbread(fin, as_version=4)
        self.logger.info(om.jobs.put(nb, name))

    def get(self):
        om = get_omega(self.args)
        local = self.args['<path>']
        name = self.args['<name>']
        notebook = om.jobs.get(name)
        nbformat.write(notebook, local)
        self.logger.debug(local)

    def metadata(self):
        om = get_omega(self.args)
        name = self.args.get('<name>')
        self.logger.info(om.jobs.metadata(name).to_json())

    def plugins(self):
        om = get_omega(self.args)
        for kind, plugincls in om.defaults.OMEGA_STORE_BACKENDS.items():
            self.logger.info(kind, plugincls.__doc__)

    def schedule(self):
        # FIXME this is a mess
        om = get_omega(self.args)
        name = self.args.get('<name>')
        at = self.args.get('--at')
        # get interval specs
        if at:
            hour, minute = at.split(':')
        else:
            hour = self.args.get('--hour')
            minute = self.args.get('--minute')
        weekday = self.args.get('--weekday')
        monthday = self.args.get('--monthday')
        month = self.args.get('--month')
        delete = self.args.get('delete')
        show = self.args.get('show')
        spec = self.args.get('--cron')
        next_n = self.args.get('--next')
        interval = self.args.get('<interval>')
        # by default we show if no interval is specified
        show = show or not any(s for s in (weekday, monthday, month, hour, minute, interval, spec))
        # print current schedule and triggers
        run_at, triggers = om.jobs.get_schedule(name, only_pending=True)
        if run_at:
            human_sched = get_description(run_at)
            self.logger.info("Currently {name} is scheduled at {human_sched}".format(**locals()))
            if next_n:
                self.logger.info("Given this existing interval, next {next_n} times would be:".format(**locals()))
                for time in om.jobs.Schedule.from_cron(run_at).next_times(int(next_n)):
                    self.logger.info("  {}".format(time))
        else:
            self.logger.info("Currently {name} is not scheduled".format(**locals()))
        # show current triggers
        if triggers:
            trigger = triggers[-1]
            if trigger['status'] == 'PENDING':
                event = trigger['event']
                self.logger.info("{name} is scheduled to run next at {event}".format(**locals()))
        # delete if currently scheduled
        if delete:
            if run_at or triggers:
                answer = self.ask("Do you want to delete this schedule?", options='Y/n', default='y')
                should_drop = answer.lower().startswith('y')
                return om.jobs.drop_schedule(name) if should_drop else None
        # create new schedule
        if not (show or delete):
            if interval:
                try:
                    # nlp text-like
                    spec = om.jobs.Schedule(interval).cron
                except Exception as e:
                    self.logger.info(f"Cannot parse {interval}, error was {e}")
                    raise
            if not spec:
                cron_repr = ('{0._orig_minute} {0._orig_hour} {0._orig_day_of_month} '
                             '{0._orig_month_of_year} {0._orig_day_of_week}')
                sched = om.jobs.Schedule(minute=minute or '*',
                                         hour=hour or '*',
                                         monthday=monthday or '*',
                                         weekday=weekday or '*',
                                         month=month or '*')
                cron_sched = sched.cron
            else:
                cron_sched = spec
            human_sched = get_description(cron_sched)
            if next_n:
                self.logger.info("Given this new interval, next {next_n} times would be:".format(**locals()))
                for time in om.jobs.Schedule.from_cron(cron_sched).next_times(int(next_n)):
                    self.logger.info("  {}".format(time))
            text = "Do you want to schedule {name} at {human_sched}?".format(**locals())
            answer = self.ask(text, options="Y/n", default='y')
            if answer.lower().startswith('n'):
                self.logger.info('Ok, not scheduled. Try again.')
                return
            self.logger.info('{name} will be scheduled to run {human_sched}'.format(**locals()))
            om.jobs.schedule(name, run_at=cron_sched, last_run=datetime.datetime.now())

    def status(self):
        om = get_omega(self.args)
        name = self.args.get('<name>')
        meta = om.jobs.metadata(name)
        attrs = meta.attributes
        runs = attrs.get('job_runs', [])
        run_at, triggers = om.jobs.get_schedule(name, only_pending=True)
        self.logger.info("Runs:")
        for run in runs:
            self.logger.info("  {ts} {status} ".format(**run))
        self.logger.info("Next scheduled runs:")
        for trigger in triggers:
            trigger['ts'] = trigger.get('ts', '')
            self.logger.info("  {ts} {status} {event-kind} {event}".format(**trigger))



# import glob
import glob
import logging
import os
import tempfile
from copy import copy
from inspect import isfunction
from zipfile import ZipFile, ZIP_DEFLATED

import dill
from mongoengine import GridFSProxy

from omegaml.backends import BaseModelBackend

ok = lambda v, vtype: isinstance(v, vtype)

logger = logging.getLogger(__name__)



class TFEstimatorModel(object):
    """
    A serializable/deserizable wrapper for a TF Estimator

    Usage:
        estimator = TFEstimatorModel(estimator_fn)
        estimator.fit(input_fn=input_fn)
        estimator.predict(input_fn=input_fn)

        The estimator_fn returns a tf.estimator.Estimator or subclass.
    """

    def __init__(self, estimator_fn, model=None, input_fn=None, model_dir=None):
        """

        Args:
            estimator_fn (func): the function to return a valid tf.estimator.Estimator instance. Called as
                                 fn(model_dir=)
            model (tf.Estimator): an existing e.g. pre-fitted Estimator instance, optional. If not specified,
                                  the model will be recreated by calling estimator_fn. If specified, the
                                  model's weights and parameters will be saved and reloaded so that a fitted
                                  model can be used without further training.
            input_fn (func|dict): the function to create the input_fn as fn(mode, X, Y, batch_size=n), where mode
                                  is either 'fit', 'evaluate', or 'predict'. If not provide defaults to an input_fn
                                  that tries to infer the correct input_fn from the method and input arguments. If
                                  provided as a dict, must contain the 'fit', 'evaluate' and 'predict' keys where
                                  each value is a valid input_fn as fn(X, Y, batch_size=n).
            model_dir (str): the model directory to use. Defaults to whatever estimator_fn/Estimator instance sets
        """
        self.estimator_fn = estimator_fn
        self._model_dir = model_dir
        self._estimator = model
        self._input_fn = input_fn

    @property
    def model_dir(self):
        return self.estimator.model_dir

    @property
    def estimator(self):
        if self._estimator is None:
            self._estimator = self.estimator_fn(model_dir=self._model_dir)
        return self._estimator

    def restore(self, model_dir):
        self._estimator = None
        self._model_dir = model_dir
        return self

    def make_input_fn(self, mode, X, Y=None, batch_size=1):
        """
        Return a tf.data.Dataset from the input provided

        Args:
            mode (str): calling mode, either 'fit', 'predict' or 'evaluate'
            X (NDArray|Tensor|Dataset): features, or Dataset of (features, labels)
            Y (NDArray|Tensor|Dataset): labels, optional

        Notes:
            X can be a Dataset of (features, labels), or just features. If X is
            just features, also provide a Dataset of just labels.

            If X, Y are NDArrays or Tensors, Dataset.from_tensor_slices((dict(X), Y))
            is used to create the Dataset. If only X is provided as a NDArray or Tensor,
            only X is used to create the Dataset.

            If none of these options work, create your own input_fn and pass it
            to the .fit/.predict methods using the input_fn= kwarg
        """
        import tensorflow as tf
        import pandas as pd
        import numpy as np

        if self._input_fn is not None:
            if isinstance(self._input_fn, dict):
                return self._input_fn[mode](X, Y=Y, batch_size=batch_size)
            else:
                return self._input_fn(mode, X, Y=Y, batch_size=batch_size)

        def input_fn():
            # if we have a dataset, use that
            if isinstance(X, tf.data.Dataset):
                if Y is None:
                    return X
                elif isinstance(Y, tf.data.Dataset):
                    return X.zip(Y)
                else:
                    return X, Y
            # if we have a dataframe, create a dataset from it
            if ok(X, pd.DataFrame) and ok(Y, pd.Series):
                dataset = tf.data.Dataset.from_tensor_slices((dict(X), Y))
                result = dataset.batch(batch_size)
            elif ok(X, pd.DataFrame):
                dataset = tf.data.Dataset.from_tensor_slices(dict(X))
                result = dataset.batch(batch_size)
            else:
                result = X, Y
            return result

        if isinstance(X, (dict, np.ndarray)):
            input_fn = tf.estimator.inputs.numpy_input_fn(x=X, y=Y, num_epochs=1, shuffle=False)
        return input_fn

    def fit(self, X=None, Y=None, input_fn=None, batch_size=100, **kwargs):
        """
        Args:
           X (Dataset|ndarray): features
           Y (Dataset|ndarray): labels, optional
        """
        assert (ok(X, object) or ok(input_fn, object)), "specify either X, Y or input_fn, not both"
        if input_fn is None:
            input_fn = self.make_input_fn('fit', X, Y, batch_size=batch_size)
        return self.estimator.train(input_fn=input_fn, **kwargs)

    def score(self, X=None, Y=None, input_fn=None, batch_size=100, **kwargs):
        """
        Args:
           X (Dataset|ndarray): features
           Y (Dataset|ndarray): labels, optional
        """
        assert (ok(X, object) or ok(input_fn, object)), "specify either X, Y or input_fn, not both"
        if input_fn is None:
            input_fn = self.make_input_fn('score', X, Y, batch_size=batch_size)
        return self.estimator.evaluate(input_fn=input_fn)

    def predict(self, X=None, Y=None, input_fn=None, batch_size=1, **kwargs):
        """
        Args:
           X (Dataset|ndarray): features
           Y (Dataset|ndarray): labels, optional
        """
        options1 = (X is None) and (input_fn is not None)
        options2 = (X is not None) and (input_fn is None)
        assert options1 or options2, "specify either X, Y or input_fn, not both"
        if input_fn is None:
            input_fn = self.make_input_fn('predict', X, Y, batch_size=batch_size)
        return self.estimator.predict(input_fn=input_fn)


class TFEstimatorModelBackend(BaseModelBackend):
    KIND = 'tfestimator.model'

    @classmethod
    def supports(self, obj, name, **kwargs):
        return isinstance(obj, TFEstimatorModel)

    def _package_model(self, obj, filename):
        model_dir = obj.model_dir
        fname = os.path.basename(filename)
        zipfname = os.path.join(self.model_store.tmppath, fname)
        # get relevant parts of model_dir
        with ZipFile(zipfname, 'w', compression=ZIP_DEFLATED) as zipf:
            zipf.writestr('modelobj.dill', dill.dumps(obj))
            for part in glob.glob(os.path.join(model_dir, '*')):
                arcname = os.path.basename(part)
                if arcname == 'modelobj.dill':
                    # ignore pre-existing model
                    continue
                zipf.write(part, arcname)
        return zipfname

    def _extract_model(self, packagefname):
        lpath = tempfile.mkdtemp()
        fname = os.path.basename(packagefname)
        mklfname = os.path.join(lpath, fname)
        with ZipFile(packagefname) as zipf:
            zipf.extractall(lpath)
        with open(os.path.join(lpath, 'modelobj.dill'), 'rb') as fin:
            model = dill.load(fin)
        model.restore(lpath)
        return model

    def put_model(self, obj, name, attributes=None):
        # create a copy so we can reset the model dir
        # this is required so the model path does not get restored on get
        obj = copy(obj)
        obj._model_dir = None
        zipfname = self._package_model(obj, name)
        with open(zipfname, 'rb') as fzip:
            fileid = self.model_store.fs.put(
                fzip, filename=self.model_store._get_obj_store_key(name, '.tfm'))
            gridfile = GridFSProxy(grid_id=fileid,
                                   db_alias='omega',
                                   collection_name=self.model_store.bucket)
        return self.model_store._make_metadata(
            name=name,
            prefix=self.model_store.prefix,
            bucket=self.model_store.bucket,
            kind=self.KIND,
            attributes=attributes,
            gridfile=gridfile).save()

    def get_model(self, name, version=-1):
        filename = self.model_store._get_obj_store_key(name, '.tfm')
        packagefname = os.path.join(self.model_store.tmppath, name)
        dirname = os.path.dirname(packagefname)
        try:
            os.makedirs(dirname)
        except OSError:
            # OSError is raised if path exists already
            pass
        outf = self.model_store.fs.get_version(filename, version=version)
        with open(packagefname, 'wb') as zipf:
            zipf.write(outf.read())
        # restore the model, this will also set the model_dir
        model = self._extract_model(packagefname)
        return model

    def fit(self, modelname, Xname, Yname=None, pure_python=True, **kwargs):
        model = self.model_store.get(modelname)
        X = self.data_store.get(Xname)
        Y = self.data_store.get(Yname) if Yname else None
        if isfunction(X) and Y is None:
            # support f
            model.fit(input_fn=X)
        else:
            model.fit(X, Y)
        meta = self.model_store.put(model, modelname)
        return meta

    def predict(
            self, modelname, Xname, rName=None, pure_python=True, **kwargs):
        import pandas as pd
        model = self.model_store.get(modelname)
        X = self.data_store.get(Xname)
        if isfunction(X):
            result = pd.DataFrame(v for v in model.predict(input_fn=X))
        else:
            result = pd.DataFrame(v for v in model.predict(X))
        if rName is not None:
            result = self.data_store.put(result, rName)
        return result

    def score(
            self, modelname, Xname, Yname, rName=True, pure_python=True,
            **kwargs):
        import pandas as pd
        model = self.model_store.get(modelname)
        X = self.data_store.get(Xname)
        Y = self.data_store.get(Yname)
        if isfunction(X) and Y is None:
            # support f
            result = model.fit(input_fn=X)
        else:
            result = model.score(X, Y)
        if not pure_python:
            result = pd.Series(result)
        if rName is not None:
            result = self.data_store.put(result, rName)
        return result



from unittest import TestCase

from tornado.web import HTTPError

from omegaml import Omega
from omegaml.notebook.omegacontentsmgr import OmegaStoreContentsManager


class OmegaContentsManagerTests(TestCase):
    def setUp(self):
        self.om = Omega()
        self.mgr = OmegaStoreContentsManager(omega=self.om)
        [self.om.jobs.drop(fn) for fn in self.om.jobs.list()]

    def _create_notebook(self, name):
        code = """
        print('hello world')
        """.strip()
        self.om.jobs.create(code, name)

    def test_get_top_level(self):
        om = self.om
        # check director is empty to start with
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 0)
        # add a new model, check it is represented
        self._create_notebook('foo')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'foo.ipynb')
        self.assertEqual(model['content'][0]['type'], 'notebook')

    def test_subdirectory(self):
        om = self.om
        # create a notebook in a sub directory
        self._create_notebook('sub/foo')
        model = self.mgr.get('/')
        # check we can see the directory
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['type'], 'directory')
        self.assertEqual(model['content'][0]['name'], 'sub')
        # check we can see the notebook
        model = self.mgr.get('/sub', type='directory')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'foo.ipynb')
        # create a notebook in a sub sub directory
        self._create_notebook('sub1/sub2/foo')
        model = self.mgr.get('/sub1', type='directory')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['type'], 'directory')
        self.assertEqual(model['content'][0]['name'], 'sub2')
        model = self.mgr.get('/sub1/sub2', type='directory')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'foo.ipynb')

    def test_save_notebook(self):
        # create a dummy notebook just so we have a valid model
        self._create_notebook('foo')
        nbmodel = self.mgr.get('foo.ipynb', type='notebook')
        # copy foo to bar
        model = self.mgr._base_model('sub', kind='directory')
        self.mgr.save(model, 'sub')
        model = self.mgr._notebook_model('foo.ipynb')
        self.mgr.save(model, 'sub/bar.ipynb')
        # check it got created within a directory
        model = self.mgr.get('/sub', type='directory')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'bar.ipynb')
        # check it can be read directly
        model = self.mgr.get('/sub/bar.ipynb', type='notebook')
        self.assertEqual(model['name'], 'bar.ipynb')
        self.assertEqual(model['content']['cells'][0]['source'], "print('hello world')")

    def test_overwrite_existing_notebook(self):
        # create a dummy notebook just so we have a valid model
        self._create_notebook('foo')
        nbmodel = self.mgr.get('foo.ipynb', type='notebook')
        # copy foo to bar
        model = self.mgr._base_model('sub', kind='directory')
        self.mgr.save(model, 'sub')
        model = self.mgr._notebook_model('foo.ipynb')
        self.mgr.save(model, 'sub/foo.ipynb')
        # save again
        model['content']['cells'][0]['source'] = "print('hello world at large!')"
        self.mgr.save(model, 'sub/foo.ipynb')
        # check it got created within a directory
        model = self.mgr.get('/sub', type='directory')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'foo.ipynb')
        # check it can be read directly
        model = self.mgr.get('/sub/foo.ipynb', type='notebook')
        self.assertEqual(model['name'], 'foo.ipynb')
        self.assertEqual(model['content']['cells'][0]['source'], "print('hello world at large!')")

    def test_create_directory(self):
        # test creating directory on top level
        model = self.mgr._base_model('sub', kind='directory')
        self.mgr.save(model, 'sub')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'sub')
        self.assertEqual(model['content'][0]['type'], 'directory')
        # test creating directory on sub level
        model = self.mgr._base_model('sub/sub1', kind='directory')
        self.mgr.save(model, 'sub/sub1')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        model = self.mgr.get('/sub', type='directory')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'sub1')
        self.assertEqual(model['content'][0]['type'], 'directory')

    def test_create_directory_escaped(self):
        # test creating directory on top level
        model = self.mgr._base_model('Untitled Folder', kind='directory')
        self.mgr.save(model, 'Untitled Folder')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'Untitled Folder')
        self.assertEqual(model['content'][0]['type'], 'directory')
        # test getting it with escaped string
        self.assertTrue(self.mgr.exists('Untitled Folder'))
        self.assertTrue(self.mgr.exists('Untitled%20Folder'))

    def test_create_notebook_in_empty_directory(self):
        # test creating directory on top level
        model = self.mgr._base_model('sub', kind='directory')
        self.mgr.save(model, 'sub')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'sub')
        self.assertEqual(model['content'][0]['type'], 'directory')
        result = self.mgr.exists('sub')
        self.assertTrue(result)
        # test creating a notebook in this new emtpty directory
        self._create_notebook('sub/foo.ipynb')
        model = self.mgr.get('sub/foo.ipynb', type='notebook')
        self.assertEqual(model['name'], 'foo.ipynb')
        self.assertEqual(model['content']['cells'][0]['source'], "print('hello world')")

    def test_create_multiple_notebook_in_empty_directory(self):
        # test creating directory on top level
        model = self.mgr._base_model('sub', kind='directory')
        self.mgr.save(model, 'sub')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'sub')
        self.assertEqual(model['content'][0]['type'], 'directory')
        result = self.mgr.exists('sub')
        self.assertTrue(result)
        # test creating a notebook in this new emtpty directory
        self._create_notebook('sub/foo.ipynb')
        self._create_notebook('sub/foo2.ipynb')
        self._create_notebook('sub/foo3.ipynb')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'sub')
        self.assertEqual(model['content'][0]['type'], 'directory')

    def test_exists(self):
        om = self.om
        # check director is empty to start with
        result = self.mgr.exists('/Untitled1')
        self.assertFalse(result)
        result = self.mgr.exists('/sub/sub1')
        self.assertFalse(result)
        # test creating directory on sub level
        model = self.mgr._base_model('sub/sub1', kind='directory')
        self.mgr.save(model, 'sub/sub1')
        result = self.mgr.exists('sub/sub1')
        self.assertTrue(result)
        # test dropping the directory actually works
        self.mgr.delete('sub/sub1')
        result = self.mgr.exists('sub/sub1')
        self.assertFalse(result)
        # test creating a file works
        model = self.mgr._base_model('Notebook.ipynb', kind='notebook')
        result = self.mgr.exists('Notebook.ipynb')

    def test_rename_notebook(self):
        # create a dummy notebook just so we have a valid model
        self._create_notebook('foo')
        nbmodel = self.mgr.get('foo.ipynb', type='notebook')
        self.mgr.rename('foo.ipynb', 'bar.ipynb')

    def test_rename_directory(self):
        model = self.mgr._base_model('sub', kind='directory')
        self.mgr.save(model, 'sub')
        self.mgr.rename_file('sub', 'new_sub')
        with self.assertRaises(HTTPError):
            model = self.mgr.get('sub', type='directory')
        model = self.mgr.get('new_sub', type='directory')
        self.assertEqual(len(model['content']), 0)



"""
omega runtime job tasks
"""
from __future__ import absolute_import

import datetime
from celery import shared_task
from celery.utils.log import get_task_logger

from omegaml.celery_util import OmegamlTask, sanitized
from omegaml.documents import MDREGISTRY


class NotebookTask(OmegamlTask):
    abstract = True

    def on_success(self, retval, task_id, *args, **kwargs):
        om = self.om
        args, kwargs = args[0:2]
        nbfile = args[0]
        meta = om.jobs.metadata(nbfile)
        attrs = meta.attributes
        attrs['state'] = 'SUCCESS'
        attrs['task_id'] = task_id
        meta.kind = MDREGISTRY.OMEGAML_JOBS

        if not kwargs:
            pass
        else:
            attrs['last_run_time'] = kwargs.get('run_at')
            attrs['next_run_time'] = kwargs.get('next_run_time')

        meta.attributes = attrs
        meta.save()

    def on_failure(self, retval, task_id, *args, **kwargs):
        om = self.om
        args, kwargs = args[0:2]
        nbfile = args[0]
        meta = om.jobs.metadata(nbfile)
        attrs = meta.attributes
        attrs['state'] = 'FAILURE'
        attrs['task_id'] = task_id
        meta.kind = MDREGISTRY.OMEGAML_JOBS

        if not kwargs:
            pass
        else:
            attrs['last_run_time'] = kwargs.get('run_at')
            attrs['next_run_time'] = kwargs.get('next_run_time')

        meta.attributes = attrs
        meta.save()


@shared_task(bind=True, base=NotebookTask)
def run_omegaml_job(self, nb_file, event=None, **kwargs):
    """
    runs omegaml job
    """
    result = self.om.jobs.run_notebook(nb_file, event=event)
    return result.to_json()


@shared_task(base=NotebookTask, bind=True)
def schedule_omegaml_job(self, nb_file, **kwargs):
    """
    schedules the running of omegaml job
    """
    result = self.om.jobs.schedule(nb_file, run_at=kwargs.get('run_at'))
    return sanitized(result)


@shared_task(base=OmegamlTask, bind=True)
def execute_scripts(self, **kwargs):
    """
    run scheduled jobs
    """
    logger = get_task_logger(self.name)
    om = self.om
    now = kwargs.get('now') or datetime.datetime.now()
    # get pending tasks, execute if time is right
    for job_meta in om.jobs.list(raw=True):
        if job_meta.name.startswith('results'):
            # ignore any scheduled results
            continue
        logger.debug("***** {}".format(job_meta))
        triggers = job_meta.attributes.get('triggers', [])
        # run pending jobs
        pending = (trigger for trigger in triggers
                   if trigger['event-kind'] == 'scheduled' and trigger['status'] == 'PENDING')
        for trigger in pending:
            run_at = trigger['run-at']
            logger.info("***** now={} run_at={}".format(now, run_at))
            if now >= run_at:
                om.runtime.job(job_meta.name).run(event=trigger['event'])
                # immediately schedule for next time
                om.jobs.schedule(job_meta.name, last_run=now)



import os
from unittest import TestCase

from omegaml import Omega
from omegaml.backends.tensorflow.protobufobj import ProtobufDataBackend
from omegaml.tests.util import OmegaTestMixin, tf_perhaps_eager_execution


# tensorflow example adopted from https://www.tensorflow.org/tutorials/load_data/tf_records

class ProtobufDataBackendTests(OmegaTestMixin, TestCase):
    def setUp(self):
        self.om = Omega()
        self.om.datasets.register_backend(ProtobufDataBackend.KIND, ProtobufDataBackend)
        self.clean()
        self._images = None
        tf_perhaps_eager_execution()

    @property
    def images(self):
        import tensorflow as tf
        if self._images is None:
            IMAGES = [
                (0,
                 'https://storage.googleapis.com/download.tensorflow.org/example_images/320px-Felis_catus-cat_on_snow.jpg')
            ]
            self._images = []
            for label, imgurl in IMAGES:
                path = tf.keras.utils.get_file(imgurl.split('/')[-1], imgurl)
                self._images.append((label, path))
        return self._images

    def create_example(self, imgpath, label):
        import tensorflow as tf
        with open(imgpath, 'rb') as fin:
            image_string = fin.read()
            image_tensor = tf.image.decode_jpeg(image_string)
        # if not in eager we need to evaluate the tensor first
        if not tf.executing_eagerly():
            with tf.Session() as sess:
                image_shape = sess.run(image_tensor).shape
        else:
            image_shape = image_tensor.shape
        # create feature record
        feature = {
            'height': _int64_feature(image_shape[0]),
            'width': _int64_feature(image_shape[1]),
            'depth': _int64_feature(image_shape[2]),
            'label': _int64_feature(label),
            'image_raw': _bytes_feature(image_string),
        }
        return tf.train.Example(features=tf.train.Features(feature=feature))

    def test_image_protobuf_get_put(self):
        import tensorflow as tf
        om = self.om
        for label, imgpath in self.images:
            example = self.create_example(imgpath, label)
            filename = os.path.basename(imgpath)
            entryname = 'images/{label}/{filename}.pbf'
            meta = om.datasets.put(example, entryname)
            self.assertEqual(meta.kind, ProtobufDataBackend.KIND)
            data = om.datasets.get(entryname)
            self.assertIsInstance(data, tf.train.Example)


# adopted from https://github.com/tensorflow/tensorflow/blob/master/tensorflow/examples/how_tos/reading_data/convert_to_records.py
def _bytes_feature(value):
    """Returns a bytes_list from a string / byte."""
    import tensorflow as tf
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))


def _float_feature(value):
    """Returns a float_list from a float / double."""
    import tensorflow as tf
    return tf.train.Feature(float_list=tf.train.FloatList(value=[value]))


def _int64_feature(value):
    """Returns an int64_list from a bool / enum / int / uint."""
    import tensorflow as tf
    return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))



from __future__ import absolute_import

import logging
from uuid import uuid4

import six

from omegaml.util import is_dataframe, settings, is_ndarray
logger = logging.getLogger(__file__)


class OmegaJobProxy(object):

    """
    proxy to a remote job in a celery worker

    Usage:

        .. code::

            om = Omega()
            # result is AsyncResult, use .get() to return it's result
            result = om.runtime.job('foojob').run()
            result.get()

            # result is AsyncResult, use .get() to return it's result
            result = om.runtime.job('foojob').schedule()
            result.get()
    """
    def __init__(self, jobname, runtime=None):
        self.jobname = jobname
        self.runtime = runtime

    def run(self, **kwargs):
        """
        run the job

        :return: the result
        """
        job_run = self.runtime.task('omegaml.notebook.tasks.run_omegaml_job')
        return job_run.delay(self.jobname, **kwargs)

    def schedule(self, **kwargs):
        """
        schedule the job
        """
        job_run = self.runtime.task('omegaml.notebook.tasks.schedule_omegaml_job')
        return job_run.delay(self.jobname, **kwargs)





def create_model():
    import numpy as np

    X = np.random.randint(0, 100, (100,)).reshape(-1, 1)
    Y = X * 2

    from sklearn.linear_model import LinearRegression

    reg = LinearRegression()
    reg.fit(X, Y)
    return reg



"""
Usage: om <command> [<action>] [<args>...] [options]
       om (models|datasets|scripts|jobs) [<args>...] [--replace] [--csv...] [options]
       om runtime [<args>...] [--async] [--result] [--param] [options]
       om cloud [<args>...] [options]
       om shell [options]
       om help [<command>]

[usage:datasets]
[usage:models]
[usage:runtime]
[usage:scripts]
[usage:jobs]
[usage:cloud]
[usage:shell]

Options:
  -h, --help          Show this screen
  --version          Show version.
  --loglevel=LEVEL   INFO,ERROR,DEBUG [default: INFO]
  --copyright        Show copyright
  --config=CONFIG    configuration file
  --bucket=BUCKET    the bucket to use
  --local-runtime    use local runtime
  -q, --noinput      don't ask for user input, assume yes
  -E                 treat patterns as regular expressions

[options:datasets]
[options:models]
[options:runtime]
[options:cloud]
[options:scripts]
[options:jobs]

[description:datasets]
[description:models]
[description:runtime]
[description:cloud]
[description:scripts]
[description:jobs]
"""
import sys

from omegaml import version
from omegaml.client.cli.catchall import CatchallCommandBase
from omegaml.client.cli.cloud import CloudCommandBase
from omegaml.client.cli.datasets import DatasetsCommandBase
from omegaml.client.cli.jobs import JobsCommandBase
from omegaml.client.cli.models import ModelsCommandBase
from omegaml.client.cli.runtime import RuntimeCommandBase
from omegaml.client.cli.scripts import ScriptsCommandBase
from omegaml.client.cli.shell import ShellCommandBase
from omegaml.client.docoptparser import CommandParser


def main(argv=None, logger=None, **kwargs):
    # make sure cli sees current project
    sys.path.insert(0, '.')
    # use argv and logger for debugging and testing
    parser = CommandParser(__doc__, [DatasetsCommandBase,
                                     ScriptsCommandBase,
                                     ModelsCommandBase,
                                     RuntimeCommandBase,
                                     CloudCommandBase,
                                     ShellCommandBase,
                                     JobsCommandBase,
                                     CatchallCommandBase],
                           argv=argv,
                           version=version,
                           logger=logger,
                           **kwargs)
    try:
        parser.parse()
        parser.process()
    except Exception as e:
        print(f"*** ERROR {e}")
        if parser.should_debug:
            raise
        exit(1)
    return parser


def climain():
    # entry point for distutils console_script
    main()


if __name__ == '__main__':
    main()



import random
from unittest import TestCase

from omegaml import Omega
import pandas as pd
from omegaml.store import Filter
from pandas.util.testing import assert_frame_equal, assert_series_equal


class MDataFrameMixinTests(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        df = self.df = pd.DataFrame({'x': list(range(0, 10)) + list(range(0, 10)),
                                     'y': list(range(0, 10)) + list(range(0, 10))})
        om = self.om = Omega()
        om.datasets.put(df, 'sample', append=False)
        self.coll = om.datasets.collection('sample')

    def test_filterops_frame(self):
        om = self.om
        df = self.df
        mdf = om.datasets.getl('sample')
        # equality selection on frame
        flt_mdf = mdf == 5
        # note MDataFrame does not actually return a mask but applies the filter
        #      to the real data. makes more sense even though the results
        #      are somewhat different this way, but match better semantically
        flt_df = df == 5  # this returns a multi-column mask with all rows(!)
        expected = df[flt_df.iloc[:, 0]]  # actual meaning in in MDataFrame
        assert_frame_equal(mdf[flt_mdf].value, expected)
        ## unequality selection
        flt_mdf = mdf < 5
        flt_df = df < 5  # this returns a multi-column mask with all rows(!)
        expected = df[flt_df.iloc[:, 0]]  # actual meaning in in MDataFrame
        assert_frame_equal(mdf[flt_mdf].value, expected)
        # combined
        flt_mdf = mdf < 5
        flt_df = df < 5  # this returns a multi-column mask with all rows(!)
        expected = df[flt_df.iloc[:, 0]]  # actual meaning in in MDataFrame
        assert_frame_equal(mdf[flt_mdf].value, expected)

    def test_filterops_series(self):
        om = self.om
        df = self.df
        mdf = om.datasets.getl('sample')
        # equality selection on frame
        flt_mdf = mdf['x'] == 5
        # note MDataFrame does not actually return a mask but applies the filter
        #      to the real data. makes more sense even though the results
        #      are somewhat different this way, but match better semantically
        flt_df = df['x'] == 5  # this returns a multi-column mask with all rows(!)
        expected = df['x'][flt_df]  # actual meaning in in MDataFrame
        assert_series_equal(mdf['x'][flt_mdf].value, expected)
        ## unequality selection
        flt_mdf = mdf < 5
        flt_df = df['x'] < 5
        expected = df['x'][flt_df]
        assert_series_equal(mdf['x'][flt_mdf].value, expected)
        # combined
        flt_mdf = mdf['x'] < 5
        flt_df = df['x'] < 5
        expected = df['x'][flt_df]  # actual meaning in in MDataFrame
        assert_series_equal(mdf['x'][flt_mdf].value, expected)

    def test_applyoperators(self):
        om = self.om
        df = self.df
        # mult
        mdf = om.datasets.getl('sample')
        mdf = mdf.apply(lambda v: v * 5)
        expected = df * 5
        assert_frame_equal(expected, mdf.value)
        # add
        mdf = om.datasets.getl('sample')
        mdf = mdf.apply(lambda v: v + 5)
        expected = df + 5
        assert_frame_equal(expected, mdf.value)
        # div
        mdf = om.datasets.getl('sample')
        mdf = mdf.apply(lambda v: v / 2)
        expected = df / 2
        assert_frame_equal(expected, mdf.value)
        # truediv
        mdf = om.datasets.getl('sample')
        mdf = mdf.apply(lambda v: v // 2)
        expected = df // 2
        assert_frame_equal(expected.astype(float), mdf.value)
        # minus
        mdf = om.datasets.getl('sample')
        mdf = mdf.apply(lambda v: v - 2)
        expected = df - 2
        assert_frame_equal(expected, mdf.value)
        # complex
        mdf = om.datasets.getl('sample')
        mdf = mdf.apply(lambda v: (v - 2) * 5)
        expected = (df - 2) * 5
        assert_frame_equal(expected, mdf.value)

    def test_apply_custom_project(self):
        om = self.om
        df = self.df

        def complexfn(ctx):
            ctx.add({
                '$project': {
                    'x': {
                        '$multiply': ['$x', 5]
                    },
                    'y': 1,
                }
            })
            return ctx

        mdf = om.datasets.getl('sample')
        mdf = mdf.apply(complexfn)
        df['x'] = df['x'] * 5
        expected = df
        # note since pandas 0.25.1 column order depends on dict key
        # insertion order, using [expected.colums] is to ensure the
        # same column order -- we can't control the key order by pymongo
        assert_frame_equal(expected, mdf.value[expected.columns])

    def test_apply_custom_project_simple(self):
        om = self.om
        df = self.df

        def complexfn(ctx):
            ctx.project(x={
                '$multiply': ['$x', 5]
            }, b={'$divide': ['$x', 5]})
            return ctx

        mdf = om.datasets.getl('sample')
        mdf = mdf.apply(complexfn)
        # simulate parallel ops
        df['x'], df['b'] = df['x'] * 5, df['x'] / 5
        value = mdf.value
        expected = df[value.columns]
        self.assertEquals(sorted(['x', 'b']), sorted(expected))
        assert_frame_equal(expected, value)

    def test_apply_dict(self):
        om = self.om

        df = pd.DataFrame({
            'x': pd.date_range(start=pd.to_datetime('09.05.2018',
                                                    format='%m.%d.%Y'), periods=5, tz=None),
            'y': range(5),
        })
        om.datasets.put(df, 'sample', append=False)

        mdf = om.datasets.getl('sample')
        mdf = mdf.apply(lambda v: dict(a=v['x'].dt.dayofweek,
                                       b=v['x'].dt.year)).value

        self.assertEqual(list(mdf['a'].values), [4, 5, 6, 7, 1])
        self.assertTrue(all(v == 2018 for v in mdf['b'].values))

    def test_apply_dt(self):
        om = self.om

        df = pd.DataFrame({
            'x': pd.date_range(start='now', periods=5, tz=None),
            'y': range(5),
        })
        om.datasets.put(df, 'sample', append=False)

        mdf = om.datasets.getl('sample')
        # FIXME we have to convert MongoDB dayofweek to Pandas dt.dayofweek
        # they start at different days of the week with different base index
        value = mdf.apply(lambda v: ((v['x'].dt.dayofweek) + 5) % 7).value
        expected = df['x'].dt.dayofweek
        assert_series_equal(expected, value['x'])

    def test_apply_str_concat(self):
        om = self.om

        df = pd.DataFrame({
            'x': ['abc', 'def'],
            'b': ['xyz', 'nop'],
        })
        om.datasets.put(df, 'sample', append=False)

        # test on dataframe
        mdf = om.datasets.getl('sample')
        mdf = mdf.apply(lambda v: v.str.concat(['*', 'b'])).value
        self.assertEqual(list(mdf['x'].values), ['abc*xyz', 'def*nop'])
        self.assertEqual(list(mdf['b'].values), ['xyz*xyz', 'nop*nop'])
        # test on series
        mdf = om.datasets.getl('sample')
        mdf = mdf['x'].apply(lambda v: v.str.concat(['foo'])).value
        self.assertEqual(list(mdf.values), ['abcfoo', 'deffoo'])

    def test_apply_str_split(self):
        om = self.om

        df = pd.DataFrame({
            'x': ['abc', 'def'],
            'b': ['xyz', 'nop'],
        })
        om.datasets.put(df, 'sample', append=False)

        # test on dataframe
        mdf = om.datasets.getl('sample')
        mdf = mdf.apply(lambda v: v.str.split('a')).value
        self.assertEqual(list(mdf['x'].values), ['', 'bc', 'def'])
        self.assertEqual(list(mdf['b'].values), ['xyz', 'xyz', 'nop'])
        # test on series
        mdf = om.datasets.getl('sample')
        mdf = mdf['x'].apply(lambda v: v.str.split(['a'])).value
        self.assertEqual(mdf.name, 'x')
        self.assertEqual(list(mdf.values), ['', 'bc', 'def'])

    def test_apply_str_lower(self):
        om = self.om

        df = pd.DataFrame({
            'x': ['ABC', 'DEF'],
            'b': ['XYZ', 'NOP'],
        })
        om.datasets.put(df, 'sample', append=False)

        # test on dataframe
        mdf = om.datasets.getl('sample')
        mdf = mdf.apply(lambda v: v.str.lower()).value
        self.assertEqual(list(mdf['x'].values), ['abc', 'def'])
        self.assertEqual(list(mdf['b'].values), ['xyz', 'nop'])
        # test on series
        mdf = om.datasets.getl('sample')
        mdf = mdf['x'].apply(lambda v: v.str.lower()).value
        self.assertEqual(list(mdf.values), ['abc', 'def'])
        self.assertEqual(mdf.name, 'x')
        mdf = om.datasets.getl('sample')
        mdf = mdf['b'].apply(lambda v: v.str.lower()).value
        self.assertEqual(list(mdf.values), ['xyz', 'nop'])
        self.assertEqual(mdf.name, 'b')

    def test_apply_str_substr(self):
        om = self.om

        df = pd.DataFrame({
            'x': ['abc', 'def'],
            'b': ['xyz', 'nop'],
        })
        om.datasets.put(df, 'sample', append=False)

        # test on dataframe
        mdf = om.datasets.getl('sample')
        mdf = mdf.apply(lambda v: v.str.substr(0, 2)).value
        self.assertEqual(list(mdf['x'].values), ['ab', 'de'])
        self.assertEqual(list(mdf['b'].values), ['xy', 'no'])
        # test on series
        mdf = om.datasets.getl('sample')
        mdf = mdf['x'].apply(lambda v: v.str.substr(0, 2)).value
        self.assertEqual(list(mdf.values), ['ab', 'de'])
        self.assertEqual(mdf.name, 'x')
        mdf = om.datasets.getl('sample')
        mdf = mdf['b'].apply(lambda v: v.str.substr(0, 2)).value
        self.assertEqual(list(mdf.values), ['xy', 'no'])
        self.assertEqual(mdf.name, 'b')

    def test_apply_str_isequal(self):
        om = self.om

        df = pd.DataFrame({
            'x': ['abc', 'def'],
            'b': ['xyz', 'nop'],
        })
        om.datasets.put(df, 'sample', append=False)

        # test on dataframe
        mdf = om.datasets.getl('sample')
        mdf = mdf.apply(lambda v: v.str.isequal('abc')).value
        self.assertEqual(list(mdf['x'].values), [True, False])
        self.assertEqual(list(mdf['b'].values), [False, False])
        # test on series
        mdf = om.datasets.getl('sample')
        mdf = mdf['x'].apply(lambda v: v.str.isequal('abc')).value
        self.assertEqual(list(mdf.values), [True, False])
        self.assertEqual(mdf.name, 'x')
        mdf = om.datasets.getl('sample')
        mdf = mdf['b'].apply(lambda v: v.str.isequal('abc')).value
        self.assertEqual(list(mdf.values), [False, False])
        self.assertEqual(mdf.name, 'b')

    def test_apply_str_len(self):
        om = self.om

        df = pd.DataFrame({
            'x': ['abc', 'def'],
            'b': ['xxyz', 'noppp'],
        })
        om.datasets.put(df, 'sample', append=False)

        # test on dataframe
        mdf = om.datasets.getl('sample')
        mdf = mdf.apply(lambda v: v.str.len()).value
        self.assertEqual(list(mdf['x'].values), [3, 3])
        self.assertEqual(list(mdf['b'].values), [4, 5])
        # test on series
        mdf = om.datasets.getl('sample')
        mdf = mdf['x'].apply(lambda v: v.str.len()).value
        self.assertEqual(list(mdf.values), [3, 3])
        self.assertEqual(mdf.name, 'x')
        mdf = om.datasets.getl('sample')
        mdf = mdf['b'].apply(lambda v: v.str.len()).value
        self.assertEqual(list(mdf.values), [4, 5])
        self.assertEqual(mdf.name, 'b')

    def test_apply_str_index(self):
        om = self.om

        df = pd.DataFrame({
            'x': ['abc', 'def'],
            'b': ['xxbc', 'bcxz'],
        })
        om.datasets.put(df, 'sample', append=False)

        # test on dataframe
        mdf = om.datasets.getl('sample')
        mdf = mdf.apply(lambda v: v.str.index('bc')).value
        self.assertEqual(list(mdf['x'].values), [1, -1])
        self.assertEqual(list(mdf['b'].values), [2, 0])
        # test on series
        mdf = om.datasets.getl('sample')
        mdf = mdf['x'].apply(lambda v: v.str.index('bc')).value
        self.assertEqual(list(mdf.values), [1, -1])
        self.assertEqual(mdf.name, 'x')
        mdf = om.datasets.getl('sample')
        mdf = mdf['b'].apply(lambda v: v.str.index('bc')).value
        self.assertEqual(list(mdf.values), [2, 0])
        self.assertEqual(mdf.name, 'b')

    def test_apply_groupby_math(self):
        om = self.om

        df = pd.DataFrame({
            'x': ['abc', 'def'] * 5,
            'v': range(7, 17),
        })

        om.datasets.put(df, 'sample', append=False)

        def groupby(ctx):
            ctx.groupby('x', v={'$sum': '$v'})['v'] * 2

        mdf = om.datasets.getl('sample')
        mdf = mdf.apply(groupby)
        expected = df.groupby('x').agg(dict(v='sum')) * 2
        assert_frame_equal(expected, mdf.value)

    def test_apply_groupby_inline(self):
        om = self.om

        df = pd.DataFrame({
            'x': ['abc', 'def'] * 5,
            'v': range(7, 17),
        })

        om.datasets.put(df, 'sample', append=False)

        mdf = om.datasets.getl('sample')
        mdf = mdf.apply(lambda v: v.groupby('x').sum('v'))
        expected = df.groupby('x').agg(dict(v='sum')).rename(columns=dict(v='v_sum'))
        assert_frame_equal(expected, mdf.value)

    def test_apply_groupby_agg(self):
        om = self.om

        df = pd.DataFrame({
            'x': ['abc', 'def'] * 5,
            'v': range(7, 17),
        })

        om.datasets.put(df, 'sample', append=False)

        mdf = om.datasets.getl('sample')
        mdf = mdf.apply(lambda v: v.groupby('x').agg(v=['sum', 'mean', 'std'])).value
        expected = df.groupby('x').agg(dict(v=['sum', 'mean', 'std']))
        self.assertEqual(list(expected[('v', 'sum')].values), list(mdf['v_sum'].values))

    def test_apply_groupby_agg(self):
        om = self.om

        df = pd.DataFrame({
            'x': ['abc', 'def'] * 5,
            'v': range(7, 17),
        })
        om.datasets.put(df, 'sample', append=False)

        mdf = om.datasets.getl('sample')
        mdf = mdf.apply(lambda v: v.groupby('x').agg(v=['sum', 'mean', 'std'])).value
        expected = df.groupby('x').agg(dict(v=['sum', 'mean', 'std']))
        self.assertEqual(list(expected[('v', 'sum')].values), list(mdf['v_sum'].values))
        self.assertEqual(list(expected[('v', 'mean')].values), list(mdf['v_avg'].values))
        self.assertEqual(list(expected[('v', 'std')].values), list(mdf['v_std'].values))

    def test_apply_cache(self):
        """
        test apply cache

        note this tests the functionality of the cache, not the performance
        """
        om = self.om

        df = pd.DataFrame({
            'x': ['abc', 'def'] * 5,
            'v': range(7, 17),
        })
        om.datasets.put(df, 'sample', append=False)

        # cache a groupby, check the cache got created
        mdf = om.datasets.getl('sample')
        cache_key = mdf.apply(lambda v: v.groupby('x').agg(v=['sum', 'mean', 'std'])).reset_cache().persist()
        cursor = mdf.apply(lambda v: v.groupby('x').agg(v=['sum', 'mean', 'std']))._get_cached_cursor()
        expected = df.groupby('x').agg(dict(v=['sum', 'mean', 'std']))
        self.assertIsNotNone(cache_key)
        self.assertIsNotNone(cursor)
        # replace the data to supersede the cache, without resetting the cache
        df = pd.DataFrame({
            'x': ['abc', 'def'] * 5,
            'v': range(17, 27),
        })
        df['v'] = df['v'] / 2
        om.datasets.put(df, 'sample', append=False)
        # evaluate the groupby, expected results are taken from cache since we did not reset the cache
        mdf = mdf.apply(lambda v: v.groupby('x').agg(v=['sum', 'mean', 'std'])).value
        self.assertEqual(list(expected[('v', 'sum')].values), list(mdf['v_sum'].values))
        self.assertEqual(list(expected[('v', 'mean')].values), list(mdf['v_avg'].values))
        self.assertEqual(list(expected[('v', 'std')].values), list(mdf['v_std'].values))
        # finally reset cache
        mdf = om.datasets.getl('sample')
        mdf = mdf.apply(lambda v: v.groupby('x').agg(v=['sum', 'mean', 'std'])).reset_cache()
        # check reset was successful
        cursor = mdf.apply(lambda v: v.groupby('x').agg(v=['sum', 'mean', 'std']))._get_cached_cursor()
        mdf = mdf.value
        self.assertIsNone(cursor)
        self.assertNotEqual(list(expected[('v', 'sum')].values), list(mdf['v_sum'].values))
        self.assertNotEqual(list(expected[('v', 'mean')].values), list(mdf['v_avg'].values))
        self.assertNotEqual(list(expected[('v', 'std')].values), list(mdf['v_std'].values))
        # try again with a full cache reset
        mdf = om.datasets.getl('sample')
        mdf.apply(lambda v: v.groupby('x').agg(v=['sum', 'mean', 'std'])).persist()
        cursor = mdf.apply(lambda v: v.groupby('x').agg(v=['sum', 'mean', 'std']))._get_cached_cursor()
        self.assertIsNotNone(cursor)
        # make sure we get a fresh mdf without any apply groups so we can test cache reset in full on collection
        mdf = om.datasets.getl('sample')
        mdf.reset_cache(full=True)
        cursor = mdf.apply(lambda v: v.groupby('x').agg(v=['sum', 'mean', 'std']))._get_cached_cursor()
        self.assertIsNone(cursor)

    def test_apply_quantile(self):
        """
        test covariance
        """
        om = self.om
        df = pd.DataFrame({
            'x': range(1000),
            'y': range(1000),
        })
        om.datasets.put(df, 'qtest', append=False)
        mdf = om.datasets.getl('qtest')
        result = mdf.quantile([.1, .2]).value
        # FIXME this is actually wrong, see df.quantile([.1, .2])
        self.assertListEqual(list(result.loc['p0.1'].values), [100, 100])
        self.assertListEqual(list(result.loc['p0.2'].values), [200, 200])

    def test_apply_covariance(self):
        """
        test covariance
        """
        om = self.om
        df = pd.DataFrame({
            'x': range(10),
            'y': range(10, 20),
        })
        om.datasets.put(df, 'covtest', append=False)
        expected = df.cov()
        mdf = om.datasets.getl('covtest')
        result = mdf.cov().value
        assert_frame_equal(result, expected)

    def test_apply_correlation(self):
        """
        test covariance
        """
        om = self.om
        df = pd.DataFrame({
            'x': range(10),
            'y': range(10, 20),
        })
        om.datasets.put(df, 'corrtest', append=False)
        expected = df.corr()
        mdf = om.datasets.getl('corrtest')
        result = mdf.corr().value
        assert_frame_equal(result, expected)


















import atexit

import logging
import signal
from pymongo import WriteConcern
from pymongo.read_concern import ReadConcern


class OmegaLoggingHandler(logging.Handler):
    """
    A Python logging handler that writes to a dataset

    You should only use this if you are familiar with Python logging
    and you are sure you need its flexibility. In most cases using
    om.logger is the better solution.

    Usage:
        # in ipython
        handler = OmegaLoggingHandler.setup(reset=True)
        handler = handler.tail()

        # if you want to attach to a specific logger
        logger = logging.getLogger(__file__)
        handler = OmegaLoggingHandler.setup(reset=True, logger=logger)

        # if omega is initialized other than om.setup()
        handler = OmegaLoggingHandler.setup(store=om.datasets)
        df = om.logger.dataset.get()

    See Also:
        OmegaSimpleLogger for notes about the log dataset format
    """

    def __init__(self, store, dataset, collection, level=None):
        super().__init__()
        level = level or logging.INFO
        self.store = store
        self.collection = collection
        self.dataset = dataset
        self.setLevel(level)

    def emit(self, record):
        log_entry = _make_record(record.levelname, record.levelno, record.name,
                                 record.msg, self.format(record))
        self.collection.insert_one(log_entry)

    def tail(self, wait=False):
        return TailableLogDataset(dataset=self.dataset, collection=self.collection).tail(wait=wait)

    @classmethod
    def setup(cls, store=None, dataset=None, level=None, logger=None,
              fmt=None, reset=False, size=10 * 1024 * 1024, defaults=None):
        """
        Args:
            dataset (str): the name of the dataset
            level (int): set any logging.INFO, logging.ERROR, logging.DEBUG
            size (int): maxium size in bytes (defaults to 1MB)
            target (Omega): omega instance to create the dataset in, defaults to the default om
            size (int): the maximum size of the log in bytes (capped), defaults to 1MB, set to -1
               for
            reset (bool): recreate the logging dataset
            fmt (str): the format specification
        """
        import omegaml as om
        import logging
        level = level or logging.INFO
        store = store or om.setup().datasets
        defaults = defaults or store.defaults
        dataset = dataset or defaults.OMEGA_LOG_DATASET
        fmt = fmt or defaults.OMEGA_LOG_FORMAT
        # setup handler and logger
        logger = logger or logging.getLogger('omegaml')
        collection = _setup_logging_dataset(store, dataset, logger=logger, size=size, reset=reset)
        formatter = logging.Formatter(fmt)
        handler = OmegaLoggingHandler(store, dataset, collection, level=level)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        return handler


class OmegaSimpleLogger:
    """
    om.logger implementation

    This is a simplified logger for client code that does not rely
    on standard python logging. Use this instead of print()

    Usage:
        # in your code, e.g. in a runtime script or in a notebook
        om.logger.info(message)
        om.logger.critical(message)
        om.logger.error(message)
        om.logger.warning(message)
        om.logger.debug(message)

        # get the log contents
        df = om.logger.dataset.get() # returns a DataFrame of the log

        # filter on specific levels
        df = om.logger.dataset.get(levelname='INFO')

        # tail the log
        om.logger.dataset.tail()

        # from the command line
        $ om runtime log
        $ om runtime log -f

    Notes:
        The log contents is written to the omegaml dataset .omega/logs. You
        should not need to interact with this dataset.

        The format of the dataframe returned by om.logger.dataset.get() is
        as follows:

            levelname (str): the log level, e.g. INFO, DEBUG, WARNING, ERROR
            levelno (int): the level as a numberic, where DEBUG is highest.
            message (str): the message as passed to the logger methods
            text (str): the formatted message, including timestamp
            created (datetime): the UTC datetime of the log message

        Tailing the log is based on querying the latest message created, and
        then subsequently returning all newer messages. The tailing is
        implemented by TailableLogDataset and runs in a separate thread. To
        run the tailing as a blocking command use om.logger.dataset.tail(wait=True).
    """
    levels = 'QUIET,CRITICAL,ERROR,WARNING,INFO,DEBUG'.split(',')

    def __init__(self, store=None, dataset=None, collection=None, level='INFO',
                 size=1 * 1024 * 1024, defaults=None):
        import omegaml as om

        self.store = store or om.setup().datasets
        self.defaults = defaults or store.defaults
        self.dsname = dataset or self.defaults.OMEGA_LOG_DATASET
        self._dataset = None
        self._level = None
        self.setLevel(level)
        self._is_setup = False
        self._collection = collection
        self.size = size

    @property
    def collection(self):
        if not self._is_setup:
            self._collection = _setup_logging_dataset(self.store, self.dsname, self,
                                                     collection=self._collection, size=self.size)
            self._is_setup = True
        return self._collection

    def setLevel(self, level):
        """
        choose a level, itself and all levels to the left of it will be logged

        Usage:
            e.g. setLevel('DEBUG') => DEBUG, INFO, WARNING, ERROR, CRITICAL
                           INFO    => INFO, WARNING, ERROR, CRITICAL
                           WARNING => WARNING, ERROR, CRITICAL
                           ERROR   => ERROR, CRITICAL
                           CRITICAL => CRITICAL

        Args:
            level (str): the level threshold

        Returns:
            None
        """
        self._level = self.levels.index(level)

    @property
    def level(self):
        return self.levels[self._level]

    def log(self, level, message):
        """
        simple logger function, use instead of print()
        """
        # check log level is below threshold (note this is reversed)
        levelno = self.levels.index(level)
        if levelno > self._level:
            return
        # insert a log message
        fmt = '{created} {level} {message}'
        log_entry = _make_record(level, levelno, 'simple', message, fmt=fmt)
        self.collection.insert_one(log_entry)

    def info(self, message, **kwargs):
        self.log('INFO', message, **kwargs)

    def error(self, message, **kwargs):
        self.log('ERROR', message, **kwargs)

    def debug(self, message, **kwargs):
        self.log('DEBUG', message, **kwargs)

    def warning(self, message, **kwargs):
        self.log('WARNING', message, **kwargs)

    def critical(self, message, **kwargs):
        self.log('CRITICAL', message, **kwargs)

    @property
    def dataset(self):
        if self._dataset is None:
            self._dataset = TailableLogDataset(self.store,
                                               dataset=self._dataset,
                                               collection=self.collection)
        return self._dataset


class TailableLogDataset:
    """
    A threaded log tailer

    Usage:
        om.logger.dataset.get()
        om.logger.dataset.tail()
    """

    def __init__(self, store, dataset=None, collection=None, stdout=None):
        logger = self
        self.store = store
        self.dataset = dataset or store.defaults.OMEGA_LOG_DATASET
        self.collection = _setup_logging_dataset(store, self.dataset, logger, collection=collection)
        self.stdout = stdout
        self.tail_thread = None

    def __repr__(self):
        return "TailableLogDataset(dataset='{self.dataset}')".format(**locals())

    def tail(self, wait=False):
        return self._start(wait=wait)

    def stop(self):
        self.tail_stop = True

    def get(self, **kwargs):
        return self.store.get(self.dataset, **kwargs)

    def _start(self, wait=False):
        from threading import Thread
        from time import sleep

        # set stdout, must be file-like, implementing .write() and .flush()
        stdout = self.stdout if self.stdout is not None else self._get_fixed_stdout()
        # start tail thread
        self.tail_thread = Thread(target=self._tailer,
                                  args=(self.collection, stdout,))
        self.tail_thread.start()
        self.tail_stop = False
        # register exit handler to stop thread
        atexit.register(self._stop_handler)
        signal.signal(signal.SIGHUP, self._stop_handler)
        if hasattr(signal, 'SIGBREAK'):
            signal.signal(signal.SIGBREAK, self._stop_handler)  # windows only
        signal.signal(signal.SIGINT, self._stop_handler)
        # block if requested
        if wait:
            while wait:
                sleep(1)
            self.tail_stop = True
        return self

    def _stop_handler(self, *args):
        self.stop()

    def _tailer(self, collection, stdout):
        # this runs in a thread to tail the log
        # adopted from https://api.mongodb.com/python/current/examples/tailable.html
        from time import sleep
        import pymongo

        def printer(record, stdout=stdout):
            print(str(record.get('text')), file=stdout, flush=True)

        first = collection.find().sort('$natural', pymongo.DESCENDING).limit(1).next()
        created = first.get('created')
        printer(first)

        while not self.tail_stop:
            cursor = collection.find({'created': {'$gt': created}})
            for record in cursor:
                printer(record)
                created = record.get('created')
                if self.tail_stop:
                    break
            sleep(.1)
        print("*** log tailing ended")

    def _get_fixed_stdout(self):
        """
        this is to support ipykernel stdout
        """
        import sys
        from ipykernel import iostream

        if isinstance(sys.stdout, iostream.OutStream):
            stdout = iostream.OutStream(sys.stdout.session,
                                        sys.stdout.pub_thread,
                                        'omega-logger')
            parent = dict(sys.stdout.parent_header)
            stdout.set_parent(parent)
        else:
            stdout = sys.stdout
        return stdout

def _make_record(level, levelno, name, message, text=None, fmt='{message}'):
    from datetime import datetime
    created = datetime.utcnow()
    text = text if text is None else fmt.format(**locals())
    return {
        'levelname': level,
        'levelno': levelno,
        'name': name,
        'msg': message,
        'created': created,
        'text': text,
    }

def _setup_logging_dataset(store, dsname, logger, collection=None, size=10 * 1024 * 1024, reset=False):
    # setup the dataset
    assert dsname, 'need a valid dsname, got {}'.format(dsname)
    if reset:
        store.drop(dsname, force=True)
    collection = collection or store.collection(dsname)
    # https://api.mongodb.com/python/current/api/pymongo/write_concern.html#pymongo.write_concern.WriteConcern
    FireAndForget = WriteConcern(w=0)
    ReadFast = ReadConcern('local')
    collection = collection.with_options(write_concern=FireAndForget, read_concern=ReadFast)
    store.put(collection, dsname)
    if collection.estimated_document_count() == 0:
        # initialize. we insert directly into the collection because the logger instance is not set up yet
        record = _make_record('SYSTEM', 999, 'system', 'log init', 'log init')
        collection.insert_one(record)
        store.mongodb.command('convertToCapped', collection.name, size=size)
    # ensure indexed
    idxs = collection.index_information()
    for idx in ('levelname', 'levelno', 'created'):
        if idx not in idxs:
            collection.create_index(idx)
    return collection



from __future__ import absolute_import

import os
from unittest import TestCase, skipUnless

from sklearn.datasets import make_classification
from sklearn.exceptions import NotFittedError
from sklearn.linear_model import LogisticRegression
from sklearn.linear_model.base import LinearRegression
from sklearn.linear_model.stochastic_gradient import SGDRegressor
from sklearn.metrics.regression import mean_squared_error
from sklearn.model_selection import GridSearchCV
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import DataConversionWarning

import numpy as np
from omegaml import Omega
from omegaml.util import delete_database, reshaped, settings
import pandas as pd
from six.moves import range


class RuntimeTests(TestCase):

    def setUp(self):
        TestCase.setUp(self)
        delete_database()

    def tearDown(self):
        TestCase.tearDown(self)

    def test_predict(self):
        # create some data
        x = np.array(list(range(0, 10)))
        y = x * 2
        df = pd.DataFrame({'x': x,
                           'y': y})
        X = df[['x']]
        Y = df[['y']]
        # put into Omega
        os.environ['DJANGO_SETTINGS_MODULE'] = ''
        om = Omega()
        om.runtime.celeryapp.conf.CELERY_ALWAYS_EAGER = True
        om.datasets.put(X, 'datax')
        om.datasets.put(Y, 'datay')
        om.datasets.get('datax')
        om.datasets.get('datay')
        # create a model locally, fit it, store in Omega
        lr = LinearRegression()
        lr.fit(X, Y)
        pred = lr.predict(X)
        om.models.put(lr, 'mymodel')
        self.assertIn('mymodel', om.models.list('*'))
        # have Omega predict it
        # -- using data already in Omega
        result = om.runtime.model('mymodel').predict('datax')
        pred1 = result.get()
        # -- using data provided locally
        #    note this is the same as
        #        om.datasets.put(X, 'foo')
        #        om.runtimes.model('mymodel').predict('foo')
        result = om.runtime.model('mymodel').predict(X)
        pred2 = result.get()
        self.assertTrue(
            (pred == pred1).all(), "runtimes prediction is different(1)")
        self.assertTrue(
            (pred == pred2).all(), "runtimes prediction is different(2)")

    def test_fit(self):
        # create some data
        x = np.array(list(range(0, 10)))
        y = x * 2
        df = pd.DataFrame({'x': x,
                           'y': y})
        X = df[['x']]
        Y = df[['y']]
        # put into Omega
        os.environ['DJANGO_SETTINGS_MODULE'] = ''
        om = Omega()
        om.runtime.celeryapp.conf.CELERY_ALWAYS_EAGER = True
        om.datasets.put(X, 'datax')
        om.datasets.put(Y, 'datay')
        om.datasets.get('datax')
        om.datasets.get('datay')
        # create a model locally, store (unfitted) in Omega
        lr = LinearRegression()
        om.models.put(lr, 'mymodel2')
        self.assertIn('mymodel2', om.models.list('*'))
        # predict locally for comparison
        lr.fit(X, Y)
        pred = lr.predict(X)
        # try predicting without fitting
        with self.assertRaises(NotFittedError):
            result = om.runtime.model('mymodel2').predict('datax')
            result.get()
        # have Omega fit the model then predict
        result = om.runtime.model('mymodel2').fit('datax', 'datay')
        result.get()
        # check the new model version metadata includes the datax/y references
        meta = om.models.metadata('mymodel2')
        self.assertIn('metaX', meta.attributes)
        self.assertIn('metaY', meta.attributes)
        # -- using data already in Omega
        result = om.runtime.model('mymodel2').predict('datax')
        pred1 = result.get()
        # -- using data provided locally
        #    note this is the same as
        #        om.datasets.put(X, 'foo')
        #        om.runtimes.model('mymodel2').predict('foo')
        result = om.runtime.model('mymodel2').fit(X, Y)
        result = om.runtime.model('mymodel2').predict(X)
        pred2 = result.get()
        # -- check the local data provided to fit was stored as intended
        meta = om.models.metadata('mymodel2')
        self.assertIn('metaX', meta.attributes)
        self.assertIn('metaY', meta.attributes)
        self.assertIn('_fitX', meta.attributes.get('metaX').get('collection'))
        self.assertIn('_fitY', meta.attributes.get('metaY').get('collection'))
        self.assertTrue(
            (pred == pred1).all(), "runtimes prediction is different(1)")
        self.assertTrue(
            (pred == pred2).all(), "runtimes prediction is different(2)")

    def test_partial_fit(self):
        # create some data
        x = np.array(list(range(0, 10)))
        y = x * 2
        df = pd.DataFrame({'x': x,
                           'y': y})
        X = df[['x']][0:2]
        Y = df[['y']][0:2]
        # put into Omega
        os.environ['DJANGO_SETTINGS_MODULE'] = ''
        om = Omega()
        om.runtime.celeryapp.conf.CELERY_ALWAYS_EAGER = True
        om.datasets.put(df[['x']], 'datax-full')
        om.datasets.put(X, 'datax')
        om.datasets.put(Y, 'datay')
        om.datasets.get('datax')
        om.datasets.get('datay')
        # create a model locally, store (unfitted) in Omega
        # -- ignore warnings on y shape
        import warnings
        warnings.filterwarnings("ignore", category=DataConversionWarning)
        lr = SGDRegressor(max_iter=1000, tol=1e-3)
        om.models.put(lr, 'mymodel2')
        # have Omega fit the model to get a start, then predict
        result = om.runtime.model('mymodel2').fit('datax', 'datay')
        result.get()
        # check the new model version metadata includes the datax/y references
        result = om.runtime.model('mymodel2').predict('datax-full')
        pred1 = result.get()
        mse = mean_squared_error(df.y, pred1)
        self.assertGreater(mse, 40)
        # fit mini batches add better training data, update model
        batch_size = 2
        for i, start in enumerate(range(0, len(df))):
            previous_mse = mse
            X = df[['x']][start:start + batch_size]
            Y = df[['y']][start:start + batch_size]
            om.datasets.put(X, 'datax-update', append=False)
            om.datasets.put(Y, 'datay-update', append=False)
            result = om.runtime.model('mymodel2').partial_fit(
                'datax-update', 'datay-update')
            result.get()
            # check the new model version metadata includes the datax/y
            # references
            result = om.runtime.model('mymodel2').predict('datax-full')
            pred1 = result.get()
            mse = mean_squared_error(df.y, pred1)
            self.assertLess(mse, previous_mse)

    def test_predict_pure_python(self):
        # create some data
        x = np.array(list(range(0, 10)))
        y = x * 2
        df = pd.DataFrame({'x': x,
                           'y': y}).astype('O')
        X = [[x] for x in list(df.x)]
        Y = [[y] for y in list(df.y)]
        # put into Omega -- assume a client with pandas, scikit learn
        os.environ['DJANGO_SETTINGS_MODULE'] = ''
        om = Omega()
        om.runtime.pure_python = True
        om.runtime.celeryapp.conf.CELERY_ALWAYS_EAGER = True
        om.datasets.put(X, 'datax')
        om.datasets.put(Y, 'datay')
        Xhat = om.datasets.get('datax')
        Yhat = om.datasets.get('datay')
        self.assertEqual([X], Xhat)
        self.assertEqual([Y], Yhat)
        # have Omega fit the model then predict
        lr = LinearRegression()
        lr.fit(X, Y)
        pred = lr.predict(X)
        om.models.put(lr, 'mymodel2')
        # -- using data provided locally
        #    note this is the same as
        #        om.datasets.put(X, 'foo')
        #        om.runtimes.model('mymodel2').predict('foo')
        result = om.runtime.model('mymodel2').predict(reshaped(X))
        pred2 = result.get()
        self.assertTrue(
            (pred == pred2).all(), "runtimes prediction is different(1)")
        self.assertTrue(
            (pred == pred2).all(), "runtimes prediction is different(2)")

    def test_predict_hdf_dataframe(self):
        # create some data
        x = np.array(list(range(0, 10)))
        y = x * 2
        df = pd.DataFrame({'x': x,
                           'y': y})
        X = df['x']
        Y = df['y']
        # put into Omega -- assume a client with pandas, scikit learn
        os.environ['DJANGO_SETTINGS_MODULE'] = ''
        om = Omega()
        om.runtime.pure_python = True
        om.runtime.celeryapp.conf.CELERY_ALWAYS_EAGER = True
        om.datasets.put(X, 'datax', as_hdf=True)
        om.datasets.put(Y, 'datay', as_hdf=True)
        # have Omega fit the model then predict
        lr = LinearRegression()
        lr.fit(reshaped(X), reshaped(Y))
        pred = lr.predict(reshaped(X))
        om.models.put(lr, 'mymodel2')
        # -- using data provided locally
        #    note this is the same as
        #        om.datasets.put(X, 'foo')
        #        om.runtimes.model('mymodel2').predict('foo')
        result = om.runtime.model('mymodel2').predict('datax')
        pred2 = result.get()
        self.assertTrue(
            (pred == pred2).all(), "runtimes prediction is different(1)")
        self.assertTrue(
            (pred == pred2).all(), "runtimes prediction is different(2)")

    def test_fit_pipeline(self):
        # create some data
        x = np.array(list(range(0, 10)))
        y = x * 2
        df = pd.DataFrame({'x': x,
                           'y': y})
        X = df[['x']]
        Y = df[['y']]
        # put into Omega
        os.environ['DJANGO_SETTINGS_MODULE'] = ''
        om = Omega()
        om.runtime.celeryapp.conf.CELERY_ALWAYS_EAGER = True
        om.datasets.put(X, 'datax')
        om.datasets.put(Y, 'datay')
        om.datasets.get('datax')
        om.datasets.get('datay')
        # create a pipeline locally, store (unfitted) in Omega
        p = Pipeline([
            ('lr', LinearRegression()),
        ])
        om.models.put(p, 'mymodel2')
        self.assertIn('mymodel2', om.models.list('*'))
        # predict locally for comparison
        p.fit(reshaped(X), reshaped(Y))
        pred = p.predict(reshaped(X))
        # have Omega fit the model then predict
        result = om.runtime.model('mymodel2').fit('datax', 'datay')
        result.get()
        result = om.runtime.model('mymodel2').predict('datax')
        pred1 = result.get()
        self.assertTrue(
            (pred == pred1).all(), "runtimes prediction is different(1)")

    def test_score(self):
        # create some data
        x = np.array(list(range(0, 10)))
        y = x * 2
        df = pd.DataFrame({'x': x,
                           'y': y})
        X = df[['x']]
        Y = df[['y']]
        # put into Omega
        os.environ['DJANGO_SETTINGS_MODULE'] = ''
        om = Omega()
        om.runtime.celeryapp.conf.CELERY_ALWAYS_EAGER = True
        om.datasets.put(X, 'datax')
        om.datasets.put(Y, 'datay')
        om.datasets.get('datax')
        om.datasets.get('datay')
        # create a model locally, fit it, store in Omega
        lr = LinearRegression()
        lr.fit(X, Y)
        scores = lr.score(X, Y)
        om.models.put(lr, 'mymodel')

    def test_gridsearch(self):
        X, y = make_classification()
        logreg = LogisticRegression(solver='liblinear')
        os.environ['DJANGO_SETTINGS_MODULE'] = ''
        om = Omega()
        om.runtime.celeryapp.conf.CELERY_ALWAYS_EAGER = True
        om.models.put(logreg, 'logreg')
        params = {
            'C': [0.1, 0.5, 1.0]
        }
        # gridsearch on runtimes
        om.runtime.model('logreg').gridsearch(X, y, parameters=params)
        meta = om.models.metadata('logreg')
        # check gridsearch was saved
        self.assertIn('gridsearch', meta.attributes)
        self.assertEqual(len(meta.attributes['gridsearch']), 1)
        self.assertIn('gsModel', meta.attributes['gridsearch'][0])
        # check we can get back the gridsearch model
        gs_model = om.models.get(meta.attributes['gridsearch'][0]['gsModel'])
        self.assertIsInstance(gs_model, GridSearchCV)

    def test_ping(self):
        om = Omega()
        result = om.runtime.ping(fox='bar')
        self.assertIn('message', result)
        self.assertIn('worker', result)
        self.assertEqual(result['kwargs'], dict(fox='bar'))





from __future__ import absolute_import
import random
from unittest.case import TestCase

from pymongo.collection import Collection

from omegaml import Omega
from omegaml.store.filtered import FilteredCollection
import pandas as pd
from six.moves import range


class FilteredCollectionTests(TestCase):

    def setUp(self):
        TestCase.setUp(self)
        df = pd.DataFrame({'x': list(range(0, 10)) + list(range(0, 10)),
                           'y': random.sample(list(range(0, 100)), 20)})
        om = Omega()
        om.datasets.put(df, 'sample', append=False)
        self.coll = om.datasets.collection('sample')

    def tearDown(self):
        TestCase.tearDown(self)

    def test_find(self):
        query = {'x': 1}
        fcoll = FilteredCollection(self.coll, query=query)
        result = list(fcoll.find())
        self.assertTrue(len(result) == 2)

    def test_find_one(self):
        query = {'x': 9}
        fcoll = FilteredCollection(self.coll, query=query)
        result = fcoll.find_one()
        self.assertIsInstance(result, dict)
        self.assertEqual(result.get('x'), 9)

    def test_find_one_and_delete(self):
        query = {'x': 9}
        fcoll = FilteredCollection(self.coll, query=query)
        result = fcoll.find_one_and_delete()
        self.assertIsInstance(result, dict)
        self.assertEqual(result.get('x'), 9)
        result = list(fcoll.find())
        self.assertEqual(len(result), 1)

    def test_find_one_and_replace(self):
        query = {'x': 9}
        fcoll = FilteredCollection(self.coll, query=query)
        result = fcoll.find_one_and_replace({'x': 9, 'xy': 9000})
        self.assertIsInstance(result, dict)
        self.assertEqual(result.get('x'), 9)
        result_n = fcoll.find_one()
        self.assertEqual(result_n.get('xy'), 9000)
        self.assertNotIn('y', result_n)

    def test_find_one_and_update(self):
        query = {'x': 9}
        fcoll = FilteredCollection(self.coll, query=query)
        result = fcoll.find_one_and_update({'$set': {'xy': 9000}})
        # make sure we get what we wanted
        self.assertIsInstance(result, dict)
        self.assertEqual(result.get('x'), 9)
        # be sure to get the same as before, then test it was updated
        result_n = fcoll.find_one({'_id': result.get('_id')})
        self.assertEqual(result_n.get('xy'), 9000)
        self.assertEqual(result_n.get('y'), result.get('y'))

    def test_count(self):
        query = {'x': 9}
        fcoll = FilteredCollection(self.coll, query=query)
        result = fcoll.count()
        self.assertEqual(result, 2)

    def test_distinct(self):
        query = {'x': 9}
        fcoll = FilteredCollection(self.coll, query=query)
        result = fcoll.distinct('x')
        self.assertEqual(result, [9])

    def test_group(self):
        query = {'x': 9}
        fcoll = FilteredCollection(self.coll, query=query)
        result = fcoll.group({'x': 1},
                             {'count': 0},
                             'function(curr, result) { result.count += 1; }')
        self.assertIsInstance(result, list)
        self.assertEqual(result[0].get('count'), 2)

    def test_map_reduce(self):
        query = {'x': 9}
        fcoll = FilteredCollection(self.coll, query=query)
        # calculate sum as a test value (values are random)
        ysum = sum(v.get('y') for v in fcoll.find())
        # use map reduce for the same sum calculation
        mapf = 'function() { emit(this.x, this.y); }'
        reducef = 'function(x, values) { return Array.sum(values); }'
        result = list(fcoll.map_reduce(mapf, reducef, 'mr_out').find())
        self.assertIsInstance(result, list)
        self.assertEqual(result[0].get('value'), ysum)

    def test_map_reduce_inline(self):
        query = {'x': 9}
        fcoll = FilteredCollection(self.coll, query=query)
        # calculate sum as a test value (values are random)
        ysum = sum(v.get('y') for v in fcoll.find())
        # use map reduce for the same sum calculation
        mapf = 'function() { emit(this.x, this.y); }'
        reducef = 'function(x, values) { return Array.sum(values); }'
        result = fcoll.inline_map_reduce(mapf, reducef)
        self.assertIsInstance(result, list)
        self.assertEqual(result[0].get('value'), ysum)
